# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "lxml==5.3.1",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the SUC3-mini NER dataset and upload it to the HF Hub."""

import bz2
import io
import re

import pandas as pd
import requests
from datasets import Split
from datasets.arrow_dataset import Dataset
from datasets.dataset_dict import DatasetDict
from huggingface_hub.hf_api import HfApi
from lxml import etree


def main() -> None:
    """Create the DaNE-mini NER dataset and uploads it to the HF Hub."""
    # Define download URLs
    url = "https://spraakbanken.gu.se/lb/resurser/meningsmangder/suc3.xml.bz2"

    # Download the data
    compressed_xml_data = requests.get(url).content

    # Decompress the data
    xml_data = bz2.decompress(compressed_xml_data)

    # Parse the XML data
    context = etree.iterparse(io.BytesIO(xml_data), events=("start", "end"))

    # Define a conversion dict to process the NER tags
    conversion_dict = dict(
        O="O",
        animal="MISC",
        event="MISC",
        inst="ORG",
        myth="MISC",
        other="MISC",
        person="PER",
        place="LOC",
        product="MISC",
        work="MISC",
    )

    # Initialise a `ner_tag` variable as the "no tag", `O`
    ner_tag = "O"

    # Iterate over the context
    records: list[dict[str, str | list[str]]] = list()
    tokens: list[str] = list()
    ner_tags: list[str] = list()
    for action, elt in context:
        # If the current element begins a name then set the `ner_tag` to the
        # corresponding `B` value
        if elt.tag == "name" and action == "start":
            ner_tag = f"B-{conversion_dict[elt.attrib['type']]}"

        # If the current element ends a name then reset the `ner_tag` to `O`
        elif elt.tag == "name" and action == "end":
            ner_tag = "O"

        # If the current element starts a word (i.e., `w`) then add the word to the
        # list of tokens, and add the current NER tag to the list of NER tags
        elif elt.tag == "w" and action == "start":
            if elt.text:
                tokens.append(elt.text)
                ner_tags.append(ner_tag)

        # If the current element ends a word then set the current NER tag to the `I`
        # version of the previous NER tag
        elif elt.tag == "w" and action == "end":
            if ner_tag.startswith("B-"):
                ner_tag = f"I-{ner_tag[2:]}"

        # If the current element ends a sentence then store all the data in the list of
        # records
        elif elt.tag == "sentence" and action == "end" and len(tokens) > 0:
            # Create document from the tokens
            doc = re.sub(" ([.,])", "\1", " ".join(tokens))

            # Double-check that there are the same number of tokens as NER tags
            assert len(tokens) == len(ner_tags)

            # Create the record and append it to the list of records
            record: dict[str, str | list[str]] = dict(
                doc=doc, tokens=tokens, ner_tags=ner_tags
            )
            records.append(record)

        # If the current element starts a sentence then reset the list of tokens and
        # NER tags, and set the current NER tag to `O`
        elif elt.tag == "sentence" and action == "start":
            tokens = list()
            ner_tags = list()
            ner_tag = "O"

    # Convert the records to a DataFrame
    df = pd.DataFrame.from_records(records)

    # Create new splits
    val_df = df.sample(n=256, random_state=4242)
    df_filtered = df[~df.index.isin(val_df.index)]
    test_df = df_filtered.sample(n=2048, random_state=4242)
    full_train_df = df_filtered[~df_filtered.index.isin(test_df.index)]
    assert isinstance(full_train_df, pd.DataFrame)
    train_df = full_train_df.sample(n=1024, random_state=4242)

    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(val_df, pd.DataFrame)
    assert isinstance(test_df, pd.DataFrame)

    # Collect datasets in a dataset dictionary
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(test_df, split=Split.TEST),
        }
    )

    # Push the dataset to the Hugging Face Hub
    dataset_id = "EuroEval/suc3-mini"
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)
    dataset.push_to_hub(dataset_id, private=True)


if __name__ == "__main__":
    main()



"""Caching utility functions."""

import typing as t
from functools import wraps

from .constants import T


def cache_arguments(
    *arguments: str, disable_condition: t.Callable[[], bool] = lambda: False
) -> t.Callable[[t.Callable[..., T]], t.Callable[..., T]]:
    """Cache specified arguments of a function.

    Args:
        arguments:
            The list of argument names to cache. If empty, all arguments are cached.
        disable_condition:
            A function that checks if cache should be disabled.

    Returns:
        A decorator that caches the specified arguments of a function.
    """

    def caching_decorator(func: t.Callable[..., T]) -> t.Callable[..., T]:
        """Decorator that caches the specified arguments of a function.

        Args:
            func:
                The function to decorate.

        Returns:
            The decorated function.
        """
        cache: dict[tuple, T] = dict()

        @wraps(func)
        def wrapper(*args, **kwargs) -> T:
            """Wrapper function that caches the specified arguments.

            Args:
                *args:
                    The positional arguments to the function.
                **kwargs:
                    The keyword arguments to the function.

            Returns:
                The result of the function.

            Raises:
                ValueError:
                    If an argument name is not found in the function parameters.
            """
            if not arguments:
                key = args + tuple(kwargs[k] for k in sorted(kwargs.keys()))
            else:
                func_params = func.__code__.co_varnames
                key_items: list[t.Any] = list()
                for arg_name in arguments:
                    if arg_name in kwargs:
                        key_items.append(kwargs[arg_name])
                    else:
                        try:
                            arg_index = func_params.index(arg_name)
                            key_items.append(args[arg_index])
                        except (ValueError, IndexError):
                            raise ValueError(
                                f"Argument {arg_name} not found in function "
                                f"{func.__name__} parameters."
                            )
                key = tuple(key_items)

            # Do not cache if the condition is met
            if key not in cache or disable_condition():
                cache[key] = func(*args, **kwargs)
            return cache[key]

        return wrapper

    return caching_decorator



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
#     "scikit-learn==1.6.1",
# ]
# ///

"""Create the Atsiliepimai sentiment dataset and upload to HF Hub."""

import pandas as pd
from datasets import Dataset, DatasetDict, load_dataset
from huggingface_hub import HfApi
from sklearn.model_selection import train_test_split


def main() -> None:
    """Create the Atsiliepimai sentiment dataset and upload to HF Hub."""
    # Define the dataset repository
    repo_id = "alexandrainst/lithuanian-sentiment-analysis"

    # Download the dataset
    dataset = load_dataset(repo_id, split="train")
    assert isinstance(dataset, Dataset)

    # Convert to dataframe
    df = dataset.to_pandas()
    assert isinstance(df, pd.DataFrame)

    # Fix columns
    df = df.rename(columns={"rating": "label"})
    df = df.loc[["text", "label"]]

    # Map labels
    label_mapping = {
        5: "positive",
        4: "neutral",
        3: "neutral",
        2: "negative",
        1: "negative",
    }
    df["label"] = df["label"].map(label_mapping)

    # Make splits
    # Dataset has 1796 samples, so we will have 1796 - 512 - 256 = 1028 samples
    # for the test split.
    val_size = 256
    train_size = 512

    # Sample final splits
    df, final_train_df = train_test_split(
        df, test_size=train_size, random_state=42, stratify=df["label"]
    )
    df, final_val_df = train_test_split(
        df, test_size=val_size, random_state=42, stratify=df["label"]
    )
    final_test_df = df

    # Reset indices
    final_train_df = final_train_df.reset_index(drop=True)
    final_val_df = final_val_df.reset_index(drop=True)
    final_test_df = final_test_df.reset_index(drop=True)

    # Create a dataset dictionary with custom splits
    dataset_dict = DatasetDict(
        {
            "train": Dataset.from_pandas(final_train_df),
            "val": Dataset.from_pandas(final_val_df),
            "test": Dataset.from_pandas(final_test_df),
        }
    )

    # Push the dataset to the Hugging Face Hub
    dataset_id = "EuroEval/atsiliepimai"
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)
    dataset_dict.push_to_hub(dataset_id, private=True)


if __name__ == "__main__":
    main()



"""Types used throughout the project."""

import collections.abc as c
import typing as t

from transformers import PreTrainedTokenizer
from transformers.trainer_utils import EvalPrediction

try:
    from transformers.tokenization_mistral_common import MistralCommonTokenizer
except ImportError:
    from transformers.tokenization_mistral_common import (
        MistralCommonBackend as MistralCommonTokenizer,
    )


if t.TYPE_CHECKING:
    from datasets.arrow_dataset import Dataset
    from numpy.typing import NDArray
    from pydantic import BaseModel

    from .data_models import BenchmarkConfig, GenerativeModelOutput

ScoreDict: t.TypeAlias = dict[str, dict[str, float] | c.Sequence[dict[str, float]]]
Predictions: t.TypeAlias = "NDArray | c.Sequence[str] | c.Sequence[c.Sequence[str]]"
Labels: t.TypeAlias = "NDArray | c.Sequence[str] | c.Sequence[c.Sequence[str]]"
Tokeniser: t.TypeAlias = PreTrainedTokenizer | MistralCommonTokenizer


class ComputeMetricsFunction(t.Protocol):
    """A function used to compute the metrics."""

    def __call__(
        self,
        model_outputs_and_labels: EvalPrediction
        | tuple[
            "NDArray | c.Sequence[str] | c.Sequence[c.Sequence[str]]",
            "NDArray | c.Sequence[str] | c.Sequence[c.Sequence[str]]",
        ],
        dataset: "Dataset",
        benchmark_config: "BenchmarkConfig",
    ) -> dict[str, float]:
        """Compute the metrics.

        Args:
            model_outputs_and_labels:
                The model outputs and labels.
            dataset:
                The dataset used for evaluation. This is only used in case any
                additional metadata is used to compute the metrics.

        Returns:
            The computed metrics.
        """
        ...


class ExtractLabelsFunction(t.Protocol):
    """A function used to extract the labels from the generated output."""

    def __call__(
        self, input_batch: dict[str, list], model_output: "GenerativeModelOutput"
    ) -> c.Sequence[str]:
        """Extract the labels from the generated output.

        Args:
            input_batch:
                The input batch.
            model_output:
                The model output.

        Returns:
            The extracted labels.
        """
        ...


class ScoringFunction(t.Protocol):
    """A function used to compute a score from a single model output."""

    def __call__(self, output: "BaseModel") -> float:
        """Compute a score from a model output.

        Args:
            output:
                A model output (Pydantic model) from the judge.

        Returns:
            A float score computed from the output.
        """
        ...


class BatchScoringFunction(t.Protocol):
    """A function used to compute batch scores from model outputs."""

    def __call__(
        self, outputs: list["BaseModel"], dataset: "Dataset | None" = None
    ) -> float:
        """Compute a batch score from model outputs.

        Args:
            outputs:
                List of model outputs (Pydantic models) from the judge.
            dataset:
                Optional dataset used for evaluation. Can be used for additional
                context when computing the score.

        Returns:
            A float score computed from the batch of outputs.
        """
        ...


def is_list_of_int(x: object) -> t.TypeGuard[c.Sequence[int]]:
    """Check if an object is a list of integers.

    Args:
        x:
            The object to check.

    Returns:
        Whether the object is a list of integers.
    """
    return isinstance(x, list) and all(isinstance(i, int) for i in x)


def is_list_of_list_of_int(x: object) -> t.TypeGuard[c.Sequence[c.Sequence[int]]]:
    """Check if an object is a list of list of integers.

    Args:
        x:
            The object to check.

    Returns:
        Whether the object is a list of list of integers.
    """
    return (
        isinstance(x, list)
        and all(isinstance(i, list) for i in x)
        and all(isinstance(j, int) for i in x for j in i)
    )


def is_list_of_str(x: object) -> t.TypeGuard[c.Sequence[str]]:
    """Check if an object is a list of integers.

    Args:
        x:
            The object to check.

    Returns:
        Whether the object is a list of strings.
    """
    return isinstance(x, list) and all(isinstance(i, str) for i in x)



"""Freshly initialised encoder models."""

import re
import typing as t
from functools import cached_property
from json import JSONDecodeError

from transformers.models.auto.configuration_auto import AutoConfig
from transformers.models.auto.tokenization_auto import AutoTokenizer
from transformers.models.electra import (
    ElectraForQuestionAnswering,
    ElectraForSequenceClassification,
    ElectraForTokenClassification,
)
from transformers.models.xlm_roberta import (
    XLMRobertaForQuestionAnswering,
    XLMRobertaForSequenceClassification,
    XLMRobertaForTokenClassification,
)

from ..data_models import ModelConfig
from ..enums import InferenceBackend, ModelType, TaskGroup
from ..exceptions import (
    InvalidBenchmark,
    InvalidModel,
    NeedsEnvironmentVariable,
    NeedsExtraInstalled,
)
from ..generation_utils import raise_if_wrong_params
from ..logging_utils import block_terminal_output
from ..types import Tokeniser
from ..utils import create_model_cache_dir, get_hf_token
from .hf import (
    HuggingFaceEncoderModel,
    align_model_and_tokeniser,
    setup_model_for_question_answering,
)

if t.TYPE_CHECKING:
    from transformers.configuration_utils import PretrainedConfig
    from transformers.modeling_utils import PreTrainedModel

    from ..data_models import BenchmarkConfig, DatasetConfig


class FreshEncoderModel(HuggingFaceEncoderModel):
    """A freshly initialised encoder model."""

    fresh_model = True
    allowed_params = {re.compile(r".*"): ["slow-tokenizer"]}

    def __init__(
        self,
        model_config: "ModelConfig",
        dataset_config: "DatasetConfig",
        benchmark_config: "BenchmarkConfig",
        log_metadata: bool = True,
    ) -> None:
        """Initialise the model.

        Args:
            model_config:
                The model configuration.
            dataset_config:
                The dataset configuration.
            benchmark_config:
                The benchmark configuration.
            log_metadata:
                Whether to log metadata about the model and the benchmark.
        """
        raise_if_wrong_params(
            model_config=model_config, allowed_params=self.allowed_params
        )

        # This is already set when calling `super.__init__`, but we need it to get a
        # value from `self.model_max_length`, so we set it here as well.
        self.model_config = model_config

        model, tokeniser = load_model_and_tokeniser(
            model_config=model_config,
            dataset_config=dataset_config,
            benchmark_config=benchmark_config,
            model_max_length=self.model_max_length,
        )
        self._model: "PreTrainedModel" = model
        self._tokeniser: Tokeniser = tokeniser

        self._model, self._tokeniser = align_model_and_tokeniser(
            model=self._model,
            tokeniser=self._tokeniser,
            model_max_length=self.model_max_length,
            raise_errors=benchmark_config.raise_errors,
        )

        # We specify `HuggingFaceEncoderModel` here instead of `VLLMModel`, as we want
        # to call the `__init__` method of the `BenchmarkModule` class.
        super(HuggingFaceEncoderModel, self).__init__(
            model_config=model_config,
            dataset_config=dataset_config,
            benchmark_config=benchmark_config,
            log_metadata=log_metadata,
        )

    @cached_property
    def num_params(self) -> int:
        """The number of parameters in the model.

        Returns:
            The number of parameters in the model.
        """
        match self.model_config.model_id:
            case "fresh-xlm-roberta-base":
                return 278_885_778
            case "fresh-electra-small":
                return 13_738_755
            case _:
                raise NotImplementedError(
                    f"Number of parameters for model {self.model_config.model_id} is "
                    "not implemented."
                )

    @cached_property
    def vocab_size(self) -> int:
        """The vocabulary size of the model.

        Returns:
            The vocabulary size of the model.
        """
        match self.model_config.model_id:
            case "fresh-xlm-roberta-base":
                return 250_002
            case "fresh-electra-small":
                return 32_000
            case _:
                raise NotImplementedError(
                    f"Vocabulary size for model {self.model_config.model_id} is not "
                    "implemented."
                )

    @cached_property
    def model_max_length(self) -> int:
        """The maximum context length of the model.

        Returns:
            The maximum context length of the model.
        """
        match self.model_config.model_id:
            case "fresh-xlm-roberta-base":
                return 512
            case "fresh-electra-small":
                return 128
            case _:
                raise NotImplementedError(
                    f"Maximum context length for model {self.model_config.model_id} is "
                    "not implemented."
                )

    @classmethod
    def model_exists(
        cls, model_id: str, benchmark_config: "BenchmarkConfig"
    ) -> bool | NeedsExtraInstalled | NeedsEnvironmentVariable:
        """Check if a model exists.

        Args:
            model_id:
                The model ID.
            benchmark_config:
                The benchmark configuration.

        Returns:
            Whether the model exists, or an error describing why we cannot check
            whether the model exists.
        """
        valid_models = ["fresh-electra-small", "fresh-xlm-roberta-base"]
        return model_id in valid_models

    @classmethod
    def get_model_config(
        cls, model_id: str, benchmark_config: "BenchmarkConfig"
    ) -> "ModelConfig":
        """Fetch the model configuration.

        Args:
            model_id:
                The model ID.
            benchmark_config:
                The benchmark configuration.

        Returns:
            The model configuration.
        """
        return ModelConfig(
            model_id=model_id,
            revision="main",
            param=None,
            task="fill-mask",
            languages=list(),
            merge=False,
            inference_backend=InferenceBackend.TRANSFORMERS,
            model_type=ModelType.ENCODER,
            fresh=True,
            model_cache_dir=create_model_cache_dir(
                cache_dir=benchmark_config.cache_dir, model_id=model_id
            ),
            adapter_base_model_id=None,
        )


def load_model_and_tokeniser(
    model_config: "ModelConfig",
    dataset_config: "DatasetConfig",
    benchmark_config: "BenchmarkConfig",
    model_max_length: int,
) -> tuple["PreTrainedModel", Tokeniser]:
    """Load the model and tokeniser.

    Args:
        model_config:
            The model configuration.
        dataset_config:
            The dataset configuration.
        benchmark_config:
            The benchmark configuration.
        model_max_length:
            The maximum context length of the model.

    Returns:
        The loaded model and tokeniser.
    """
    config: "PretrainedConfig"
    block_terminal_output()

    # Get the fresh model ID and the corresponding real model ID
    model_id = model_config.model_id.replace("-", "_")
    fresh_to_real_model_id_mapping = dict(
        fresh_xlm_roberta_base="FacebookAI/xlm-roberta-base",
        fresh_electra_small="google/electra-small-discriminator",
    )
    real_model_id = fresh_to_real_model_id_mapping[model_id]

    match dataset_config.task.task_group:
        case (
            TaskGroup.SEQUENCE_CLASSIFICATION
            | TaskGroup.MULTIPLE_CHOICE_CLASSIFICATION
            | TaskGroup.SPEED
        ):
            model_cls_mapping = dict(
                fresh_xlm_roberta_base=XLMRobertaForSequenceClassification,
                fresh_electra_small=ElectraForSequenceClassification,
            )
        case TaskGroup.TOKEN_CLASSIFICATION:
            model_cls_mapping = dict(
                fresh_xlm_roberta_base=XLMRobertaForTokenClassification,
                fresh_electra_small=ElectraForTokenClassification,
            )
        case TaskGroup.QUESTION_ANSWERING:
            model_cls_mapping = dict(
                fresh_xlm_roberta_base=XLMRobertaForQuestionAnswering,
                fresh_electra_small=ElectraForQuestionAnswering,
            )
        case _:
            raise InvalidBenchmark(
                f"Task group {dataset_config.task.task_group} is not "
                f"supported for model {model_config.model_id}."
            )
    model_cls = model_cls_mapping[model_id]

    # Special case where there is a mismatch between the labels during training and
    # testing
    if dataset_config.task.task_group == TaskGroup.MULTIPLE_CHOICE_CLASSIFICATION:
        id2label = {0: "0", 1: "1"}
    else:
        id2label = dataset_config.id2label

    config = AutoConfig.from_pretrained(
        real_model_id,
        token=get_hf_token(api_key=benchmark_config.api_key),
        num_labels=len(id2label),
        id2label=id2label,
        label2id={label: id_ for id_, label in id2label.items()},
        cache_dir=model_config.model_cache_dir,
        trust_remote_code=benchmark_config.trust_remote_code,
    )
    model = model_cls(config)

    if dataset_config.task.task_group == TaskGroup.QUESTION_ANSWERING:
        model = setup_model_for_question_answering(model=model)

    # Load the tokeniser. If the model is a subclass of a RoBERTa model then we
    # have to add a prefix space to the tokens, by the way the model is constructed
    prefix_models = ["Roberta", "GPT", "Deberta"]
    prefix = any(model_type in type(model).__name__ for model_type in prefix_models)
    try:
        tokeniser: Tokeniser = AutoTokenizer.from_pretrained(
            real_model_id,
            revision=model_config.revision,
            token=get_hf_token(api_key=benchmark_config.api_key),
            add_prefix_space=prefix,
            cache_dir=model_config.model_cache_dir,
            use_fast=False if model_config.param == "slow-tokenizer" else True,
            verbose=False,
            trust_remote_code=benchmark_config.trust_remote_code,
        )
    except (JSONDecodeError, OSError) as e:
        raise InvalidModel(
            f"Could not load tokeniser for model {real_model_id!r}."
        ) from e

    model, tokeniser = align_model_and_tokeniser(
        model=model,
        tokeniser=tokeniser,
        model_max_length=model_max_length,
        raise_errors=benchmark_config.raise_errors,
    )

    return model, tokeniser



"""Class that benchmarks language models."""

import collections.abc as c
import contextlib
import datetime as dt
import json
import logging
import os
import re
import typing as t
from pathlib import Path
from shutil import rmtree
from time import sleep

from torch.distributed import destroy_process_group

from .benchmark_config_factory import build_benchmark_config
from .constants import ATTENTION_BACKENDS, GENERATIVE_PIPELINE_TAGS
from .data_loading import load_data, load_raw_data
from .data_models import BenchmarkConfigParams, BenchmarkResult
from .dataset_configs import get_all_dataset_configs
from .enums import Device, GenerativeType, ModelType
from .exceptions import HuggingFaceHubDown, InvalidBenchmark, InvalidModel
from .finetuning import finetune
from .generation import generate
from .logging_utils import adjust_logging_level, get_pbar, log, log_once
from .model_config import get_model_config
from .model_loading import load_model
from .scores import log_scores
from .speed_benchmark import benchmark_speed
from .tasks import SPEED
from .utils import (
    enforce_reproducibility,
    internet_connection_available,
    split_model_id,
)

if t.TYPE_CHECKING:
    from .benchmark_modules import BenchmarkModule
    from .data_models import BenchmarkConfig, DatasetConfig, ModelConfig, Task


class Benchmarker:
    """Benchmarking all the language models.

    Attributes:
        benchmark_config_default_params:
            The default parameters for the benchmark configuration.
        benchmark_config:
            The benchmark configuration.
        force:
            Whether to force evaluations of models, even if they have been benchmarked
            already.
        results_path:
            The path to the results file.
        benchmark_results:
            The benchmark results.
    """

    def __init__(
        self,
        progress_bar: bool = True,
        save_results: bool = True,
        task: "str | Task | c.Sequence[str | Task] | None" = None,
        dataset: "str | DatasetConfig | c.Sequence[str | DatasetConfig] | None" = None,
        language: str | c.Sequence[str] = "all",
        device: Device | None = None,
        finetuning_batch_size: int = 32,
        raise_errors: bool = False,
        cache_dir: str = ".euroeval_cache",
        api_key: str | None = None,
        force: bool = False,
        verbose: bool = False,
        trust_remote_code: bool = False,
        clear_model_cache: bool = False,
        evaluate_test_split: bool = False,
        few_shot: bool = True,
        num_iterations: int = 10,
        api_base: str | None = None,
        api_version: str | None = None,
        gpu_memory_utilization: float = 0.8,
        attention_backend: str = "FLASHINFER",
        generative_type: GenerativeType | None = None,
        custom_datasets_file: Path | str = Path("custom_datasets.py"),
        debug: bool = False,
        run_with_cli: bool = False,
        requires_safetensors: bool = False,
        download_only: bool = False,
        model_language: str | c.Sequence[str] | None = None,
        dataset_language: str | c.Sequence[str] | None = None,
        batch_size: int | None = None,
    ) -> None:
        """Initialise the benchmarker.

        Args:
            progress_bar:
                Whether progress bars should be shown. Defaults to True.
            save_results:
                Whether to save the benchmark results to
                'euroeval_benchmark_results.jsonl'. Defaults to True.
            task:
                The tasks benchmark the model(s) on. Mutually exclusive with `dataset`.
                If both `task` and `dataset` are None then all datasets will be
                benchmarked.
            dataset:
                The datasets to benchmark on. Mutually exclusive with `task`. If both
                `task` and `dataset` are None then all datasets will be benchmarked.
            language:
                The language codes of the languages to include, both for models and
                datasets. Set this to 'all' if all languages should be considered.
                Defaults to "all".
            device:
                The device to use for benchmarking. Defaults to None.
            finetuning_batch_size:
                The batch size to use when finetuning. Defaults to 32.
            raise_errors:
                Whether to raise errors instead of skipping the model evaluation.
                Defaults to False.
            cache_dir:
                Directory to store cached models. Defaults to '.euroeval_cache'.
            api_key:
                The API key to use for a given inference API.
            force:
                Whether to force evaluations of models, even if they have been
                benchmarked already. Defaults to False.
            verbose:
                Whether to output additional output. This is automatically set if
                `debug` is True. Defaults to False.
            trust_remote_code:
                Whether to trust remote code when loading models. Defaults to False.
            clear_model_cache:
                Whether to clear the model cache after benchmarking each model.
                Defaults to False.
            evaluate_test_split:
                Whether to evaluate the test split of the datasets. Defaults to False.
            few_shot:
                Whether to only evaluate the model using few-shot evaluation. Only
                relevant if the model is generative. Defaults to True.
            num_iterations:
                The number of times each model should be evaluated. This is only meant
                to be used for power users, and scores will not be allowed on the
                leaderboards if this is changed. Defaults to 10.
            api_base:
                The base URL for a given inference API. Only relevant if `model` refers
                to a model on an inference API. Defaults to None.
            api_version:
                The version of the API to use. Defaults to None.
            gpu_memory_utilization:
                The GPU memory utilization to use for vLLM. Only relevant if the model
                is generative. A larger value will result in faster evaluation, but at
                the risk of running out of GPU memory. Only reduce this if you are
                running out of GPU memory. Defaults to 0.9.
            attention_backend:
                The attention backend to use for vLLM. Defaults to FLASHINFER. Only
                relevant if the model is generative.
            generative_type:
                The type of generative model to benchmark. Only relevant if the model is
                generative. If not specified, then the type will be inferred based on
                the tags of the model. Defaults to None.
            custom_datasets_file:
                Path to a Python file defining custom datasets. Defaults to
                'custom_datasets.py'.
            debug:
                Whether to output debug information. Defaults to False.
            run_with_cli:
                Whether the benchmarker is being run from the command-line interface.
                Defaults to False.
            requires_safetensors:
                Whether to only allow models that use the safetensors format. Defaults
                to False.
            download_only:
                Whether to only download models and datasets without performing any
                benchmarking. Defaults to False.
            model_language:
                Deprecated argument. Please use `language` instead.
            dataset_language:
                Deprecated argument. Please use `language` instead.
            batch_size:
                Deprecated argument. Please use `finetuning_batch_size` instead.

        Raises:
            ValueError:
                If both `task` and `dataset` are specified, or if `download_only`
                is True and we have no internet connection.
            ImportError:
                If `hf_transfer` is enabled but not installed.
        """
        if task is not None and dataset is not None:
            raise ValueError("Only one of `task` and `dataset` can be specified.")

        if not internet_connection_available() and download_only:
            msg = "It appears you do not have an internet connection, but "
            if run_with_cli:
                msg += "the --download-only flag was set."
            else:
                msg += "the argument `download_only` was set to True."
            raise ValueError(msg)

        # Deprecation warnings
        if batch_size is not None:
            if run_with_cli:
                msg = (
                    "The --batch-size option is deprecated and will be removed in a "
                    "future version. Please use --finetuning-batch-size instead. "
                    "Overwriting --finetuning-batch-size with the value from "
                    "--batch-size."
                )
            else:
                msg = (
                    "The `batch_size` argument is deprecated and will be removed in a "
                    "future version. Please use `finetuning_batch_size` instead. "
                    "Overwriting `finetuning_batch_size` with the value from "
                    "`batch_size`."
                )
            log(msg, level=logging.WARNING)
            finetuning_batch_size = batch_size
        if model_language is not None:
            if run_with_cli:
                msg = (
                    "The --model-language option is deprecated and will be removed in "
                    "a future version. Please use --language instead. Ignoring the "
                    "--model-language value."
                )
            else:
                msg = (
                    "The `model_language` argument is deprecated and will be removed "
                    "in a future version. Please use `language` instead. Ignoring the "
                    "`model_language` value."
                )
            log(msg, level=logging.WARNING)
        if dataset_language is not None:
            if run_with_cli:
                msg = (
                    "The --dataset-language option is deprecated and will be removed "
                    "in a future version. Please use --language instead. Ignoring the "
                    "--dataset-language value."
                )
            else:
                msg = (
                    "The `dataset_language` argument is deprecated and will be removed "
                    "in a future version. Please use `language` instead. Ignoring the "
                    "`dataset_language` value."
                )
            log(msg, level=logging.WARNING)

        # If FULL_LOG has been set, then force verbose mode
        if os.getenv("FULL_LOG", "0") == "1":
            verbose = True

        self.benchmark_config_default_params = BenchmarkConfigParams(
            task=task,
            dataset=dataset,
            progress_bar=progress_bar,
            save_results=save_results,
            language=language,
            device=device,
            finetuning_batch_size=finetuning_batch_size,
            raise_errors=raise_errors,
            cache_dir=cache_dir,
            api_key=api_key,
            api_base=api_base,
            api_version=api_version,
            trust_remote_code=trust_remote_code,
            clear_model_cache=clear_model_cache,
            evaluate_test_split=evaluate_test_split,
            few_shot=few_shot,
            num_iterations=num_iterations,
            requires_safetensors=requires_safetensors,
            download_only=download_only,
            gpu_memory_utilization=gpu_memory_utilization,
            attention_backend=attention_backend,
            generative_type=generative_type,
            custom_datasets_file=Path(custom_datasets_file),
            verbose=verbose,
            force=force,
            debug=debug,
            run_with_cli=run_with_cli,
        )

        self.benchmark_config = build_benchmark_config(
            benchmark_config_params=self.benchmark_config_default_params
        )

        # Initialise variable storing model lists, so we only have to fetch it once
        self._model_lists: dict[str, c.Sequence[str]] | None = None

        self.results_path = Path.cwd() / "euroeval_benchmark_results.jsonl"
        adjust_logging_level(verbose=self.benchmark_config.verbose)

    @property
    def benchmark_results(self) -> c.Sequence[BenchmarkResult]:
        """The benchmark results.

        Returns:
            A list of benchmark results.

        Raises:
            ValueError:
                If there is an error decoding a line in the results file.
        """
        if self.results_path.exists():
            benchmark_results: list[BenchmarkResult] = list()
            with self.results_path.open() as f:
                for line in f:
                    if line.strip():
                        try:
                            result_dict = json.loads(line.strip())
                        except json.JSONDecodeError as e:
                            raise ValueError(
                                f"Error decoding JSON line: {line.strip()}"
                            ) from e

                        # Fix for older records
                        has_old_raw_results = (
                            "results" in result_dict
                            and isinstance(result_dict["results"], dict)
                            and "raw" in result_dict["results"]
                            and isinstance(result_dict["results"]["raw"], dict)
                            and "test" in result_dict["results"]["raw"]
                        )
                        if has_old_raw_results:
                            result_dict["results"]["raw"] = result_dict["results"][
                                "raw"
                            ]["test"]

                        result = BenchmarkResult.from_dict(result_dict)
                        benchmark_results.append(result)
            return benchmark_results
        else:
            return list()

    def _download(
        self,
        dataset_config: "DatasetConfig",
        model_config: "ModelConfig",
        benchmark_config: "BenchmarkConfig",
    ) -> None:
        """Download data, metrics, and model for the given dataset, and model.

        Args:
            dataset_config: The configuration for the dataset.
            model_config: The configuration for the model.
            benchmark_config: The configuration for the benchmark.
        """
        log_once(
            f"Loading data for {dataset_config.logging_string}", level=logging.INFO
        )
        dataset = load_raw_data(
            dataset_config=dataset_config, cache_dir=benchmark_config.cache_dir
        )
        del dataset

        model = load_model(
            model_config=model_config,
            dataset_config=dataset_config,
            benchmark_config=benchmark_config,
        )
        del model

        log_once(
            f"Loading metrics for the '{dataset_config.task.name}' task",
            level=logging.INFO,
        )
        for metric_name in dataset_config.task.metrics:
            log_once(f"Loading metric {metric_name.name}", level=logging.DEBUG)
            metric = metric_name.download(cache_dir=benchmark_config.cache_dir)
            del metric

    def benchmark(
        self,
        model: c.Sequence[str] | str,
        task: "str | Task | c.Sequence[str | Task] | None" = None,
        dataset: "str | DatasetConfig | c.Sequence[str | DatasetConfig] | None" = None,
        progress_bar: bool | None = None,
        save_results: bool | None = None,
        language: str | c.Sequence[str] | None = None,
        device: Device | None = None,
        finetuning_batch_size: int | None = None,
        raise_errors: bool | None = None,
        cache_dir: str | None = None,
        api_key: str | None = None,
        api_base: str | None = None,
        api_version: str | None = None,
        trust_remote_code: bool | None = None,
        clear_model_cache: bool | None = None,
        evaluate_test_split: bool | None = None,
        few_shot: bool | None = None,
        num_iterations: int | None = None,
        requires_safetensors: bool | None = None,
        download_only: bool | None = None,
        gpu_memory_utilization: float | None = None,
        generative_type: GenerativeType | None = None,
        attention_backend: t.Literal[
            *ATTENTION_BACKENDS  # pyrefly: ignore[invalid-literal]
        ]
        | None = None,
        custom_datasets_file: Path | str | None = None,
        force: bool | None = None,
        verbose: bool | None = None,
        debug: bool | None = None,
        model_language: str | c.Sequence[str] | None = None,
        dataset_language: str | c.Sequence[str] | None = None,
        batch_size: int | None = None,
    ) -> c.Sequence[BenchmarkResult]:
        """Benchmarks models on datasets.

        Args:
            model:
                The full Hugging Face Hub path(s) to the pretrained transformer model.
                The specific model version to use can be added after the suffix '@':
                "model@v1.0.0". It can be a branch name, a tag name, or a commit id,
                and defaults to the latest version if not specified.
            task:
                The tasks benchmark the model(s) on. Mutually exclusive with `dataset`.
                If both `task` and `dataset` are None then all datasets will be
                benchmarked. Defaults to None.
            dataset:
                The datasets to benchmark on. Mutually exclusive with `task`. If both
                `task` and `dataset` are None then all datasets will be benchmarked.
                Defaults to None.
            progress_bar:
                Whether progress bars should be shown. Defaults to the value specified
                when initialising the benchmarker.
            save_results:
                Whether to save the benchmark results to
                'euroeval_benchmark_results.jsonl'. Defaults to the value specified
                when initialising the benchmarker.
            language:
                The language codes of the languages to include, both for models and
                datasets. Here 'no' means both Bokmål (nb) and Nynorsk (nn). Set this to
                'all' if all languages should be considered. Defaults to the value
                specified when initialising the benchmarker.
            device:
                The device to use for benchmarking. Defaults to the value specified when
                initialising the benchmarker.
            finetuning_batch_size:
                The batch size to use for finetuning. Defaults to the value specified
                when initialising the benchmarker.
            raise_errors:
                Whether to raise errors instead of skipping the model evaluation.
            cache_dir:
                Directory to store cached models. Defaults to the value specified when
                initialising the benchmarker.
            api_key:
                The API key to use for a given inference server. Defaults to the value
                specified when initialising the benchmarker.
            api_base:
                The base URL for a given inference API. Only relevant if `model` refers
                to a model on an inference API. Defaults to the value specified when
                initialising the benchmarker.
            api_version:
                The version of the API to use. Defaults to the value specified when
                initialising the benchmarker.
            trust_remote_code:
                Whether to trust remote code when loading models. Defaults to the value
                specified when initialising the benchmarker.
            clear_model_cache:
                Whether to clear the model cache after benchmarking each model. Defaults
                to the value specified when initialising the benchmarker.
            evaluate_test_split:
                Whether to evaluate the test split of the datasets. Defaults to the
                value specified when initialising the benchmarker.
            few_shot:
                Whether to only evaluate the model using few-shot evaluation. Only
                relevant if the model is generative. Defaults to the value specified
                when initialising the benchmarker.
            num_iterations:
                The number of times each model should be evaluated. This is only meant
                to be used for power users, and scores will not be allowed on the
                leaderboards if this is changed. Defaults to the value specified when
                initialising the benchmarker.
            requires_safetensors:
                Whether to only allow models that use the safetensors format. Defaults
                to the value specified when initialising the benchmarker.
            download_only:
                Whether to only download the models without evaluating them. Defaults
                to the value specified when initialising the benchmarker.
            gpu_memory_utilization:
                The GPU memory utilization to use for vLLM. Only relevant if the model
                is generative. A larger value will result in faster evaluation, but at
                the risk of running out of GPU memory. Only reduce this if you are
                running out of GPU memory. Defaults to the value specified when
                initialising the benchmarker.
            generative_type:
                The type of generative model to benchmark. Only relevant if the model is
                generative. If not specified, then the type will be inferred based on
                the tags of the model. Defaults to the value specified when initialising
                the benchmarker.
            custom_datasets_file:
                Path to a Python file defining custom datasets. Defaults to the value
                specified when initialising the benchmarker.
            force:
                Whether to force evaluations of models, even if they have been
                benchmarked already. Defaults to the value specified when initialising
                the benchmarker.
            verbose:
                Whether to output additional output. Defaults to the value specified
                when initialising the benchmarker.
            debug:
                Whether to output debug information. Defaults to the value specified
                when initialising the benchmarker.
            model_language:
                Deprecated argument. Please use `language` instead.
            dataset_language:
                Deprecated argument. Please use `language` instead.
            batch_size:
                Deprecated argument. Please use `finetuning_batch_size` instead.

        Returns:
            A list of benchmark results.

        Raises:
            ValueError:
                If both `task` and `dataset` are specified.
        """
        if task is not None and dataset is not None:
            raise ValueError("Only one of `task` and `dataset` can be specified.")

        # Deprecation warnings
        if batch_size is not None:
            log(
                "The `batch_size` argument is deprecated and will be removed in a "
                "future version. Please use `finetuning_batch_size` instead. "
                "Overwriting `finetuning_batch_size` with the value from "
                "`batch_size`.",
                level=logging.WARNING,
            )
            finetuning_batch_size = batch_size
        if model_language is not None:
            log(
                "The `model_language` argument is deprecated and will be removed "
                "in a future version. Please use `language` instead. Ignoring the "
                "`model_language` value.",
                level=logging.WARNING,
            )
        if dataset_language is not None:
            log(
                "The `dataset_language` argument is deprecated and will be removed "
                "in a future version. Please use `language` instead. Ignoring the "
                "`dataset_language` value.",
                level=logging.WARNING,
            )

        # Get a new updated benchmark configuration, based on any changes to the
        # parameters
        benchmark_config_params = BenchmarkConfigParams(
            task=(
                task if task is not None else self.benchmark_config_default_params.task
            ),
            dataset=(
                dataset
                if dataset is not None
                else self.benchmark_config_default_params.dataset
            ),
            progress_bar=(
                progress_bar
                if progress_bar is not None
                else self.benchmark_config_default_params.progress_bar
            ),
            save_results=(
                save_results
                if save_results is not None
                else self.benchmark_config_default_params.save_results
            ),
            language=(
                language
                if language is not None
                else self.benchmark_config_default_params.language
            ),
            device=(
                device
                if device is not None
                else self.benchmark_config_default_params.device
            ),
            finetuning_batch_size=(
                finetuning_batch_size
                if finetuning_batch_size is not None
                else self.benchmark_config_default_params.finetuning_batch_size
            ),
            raise_errors=(
                raise_errors
                if raise_errors is not None
                else self.benchmark_config_default_params.raise_errors
            ),
            cache_dir=(
                cache_dir
                if cache_dir is not None
                else self.benchmark_config_default_params.cache_dir
            ),
            api_key=(
                api_key
                if api_key is not None
                else self.benchmark_config_default_params.api_key
            ),
            api_base=(
                api_base
                if api_base is not None
                else self.benchmark_config_default_params.api_base
            ),
            api_version=(
                api_version
                if api_version is not None
                else self.benchmark_config_default_params.api_version
            ),
            trust_remote_code=(
                trust_remote_code
                if trust_remote_code is not None
                else self.benchmark_config_default_params.trust_remote_code
            ),
            clear_model_cache=(
                clear_model_cache
                if clear_model_cache is not None
                else self.benchmark_config_default_params.clear_model_cache
            ),
            evaluate_test_split=(
                evaluate_test_split
                if evaluate_test_split is not None
                else self.benchmark_config_default_params.evaluate_test_split
            ),
            few_shot=(
                few_shot
                if few_shot is not None
                else self.benchmark_config_default_params.few_shot
            ),
            num_iterations=(
                num_iterations
                if num_iterations is not None
                else self.benchmark_config_default_params.num_iterations
            ),
            requires_safetensors=(
                requires_safetensors
                if requires_safetensors is not None
                else self.benchmark_config_default_params.requires_safetensors
            ),
            download_only=(
                download_only
                if download_only is not None
                else self.benchmark_config_default_params.download_only
            ),
            gpu_memory_utilization=(
                gpu_memory_utilization
                if gpu_memory_utilization is not None
                else self.benchmark_config_default_params.gpu_memory_utilization
            ),
            generative_type=(
                generative_type
                if generative_type is not None
                else self.benchmark_config_default_params.generative_type
            ),
            attention_backend=(
                attention_backend
                if attention_backend is not None
                else self.benchmark_config_default_params.attention_backend
            ),
            custom_datasets_file=(
                Path(custom_datasets_file)
                if custom_datasets_file is not None
                else self.benchmark_config_default_params.custom_datasets_file
            ),
            force=(
                force
                if force is not None
                else self.benchmark_config_default_params.force
            ),
            verbose=(
                verbose
                if verbose is not None
                else self.benchmark_config_default_params.verbose
            ),
            debug=(
                debug
                if debug is not None
                else self.benchmark_config_default_params.debug
            ),
            run_with_cli=self.benchmark_config_default_params.run_with_cli,
        )
        benchmark_config = build_benchmark_config(
            benchmark_config_params=benchmark_config_params
        )

        adjust_logging_level(verbose=benchmark_config.verbose)

        if benchmark_config.clear_model_cache:
            clear_model_cache_fn(cache_dir=benchmark_config.cache_dir)

        model_ids = self._prepare_model_ids(model_id=model)
        dataset_configs = benchmark_config.datasets

        # Get all the model configs
        model_configs: list["ModelConfig"] = list()
        for model_id in get_pbar(
            iterable=model_ids,
            desc="Fetching model configurations",
            disable=not benchmark_config.verbose or not benchmark_config.progress_bar,
        ):
            try:
                model_config = get_model_config(
                    model_id=model_id, benchmark_config=benchmark_config
                )
                model_configs.append(model_config)
            except InvalidModel as e:
                log(e.message, level=logging.ERROR)

        # Create a dictionary that takes each model config to the dataset configs that
        # we need to benchmark the model on. We initially include all the relevant
        # datasets for each model.
        model_config_to_dataset_configs: dict[
            "ModelConfig", c.Sequence["DatasetConfig"]
        ] = {
            model_config: [
                dataset_config
                for dataset_config in dataset_configs
                if model_config.model_type in dataset_config.allowed_model_types
            ]
            for model_config in model_configs
        }

        # Initialise the current benchmark results with all the ones that we have cached
        # on disk already (can be none), and remove those datasets from the mapping
        current_benchmark_results: list[BenchmarkResult] = list()
        for (
            model_config,
            model_dataset_configs,
        ) in model_config_to_dataset_configs.items():
            new_model_dataset_configs: list["DatasetConfig"] = list()
            for dataset_config in model_dataset_configs:
                benchmark_record = get_record(
                    model_config=model_config,
                    dataset_config=dataset_config,
                    benchmark_config=benchmark_config,
                    benchmark_results=self.benchmark_results,
                )
                if benchmark_record is not None and not benchmark_config.force:
                    current_benchmark_results.append(benchmark_record)
                else:
                    new_model_dataset_configs.append(dataset_config)
            model_config_to_dataset_configs[model_config] = new_model_dataset_configs

        total_benchmarks = sum(
            len(dataset_configs)
            for dataset_configs in model_config_to_dataset_configs.values()
        )
        if total_benchmarks == 0:
            log(
                "No benchmarks to run, as all the selected models have already been "
                "benchmarked on all the selected datasets.",
                level=logging.INFO,
            )
            return current_benchmark_results

        num_finished_benchmarks = 0
        benchmark_params_to_revert: dict[str, t.Any] = dict()
        for model_config in model_configs:
            if not model_config_to_dataset_configs[model_config]:
                log(
                    f"Skipping model {model_config.model_id!r} because it has "
                    "already been benchmarked on all valid datasets.",
                    level=logging.DEBUG,
                )
                continue

            if model_config.adapter_base_model_id:
                open_issue_msg = (
                    "If offline support is important to you, please consider opening "
                    "an issue at https://github.com/EuroEval/EuroEval/issues."
                )
                if not internet_connection_available():
                    raise InvalidModel(
                        "Offline benchmarking of models with adapters is not currently "
                        "supported. An active internet connection is required. "
                        "{open_issue_msg}"
                    )
                elif benchmark_config.download_only:
                    log_once(
                        "You are using download only mode with a model that includes "
                        "an adapter. Please note that offline benchmarking of "
                        "adapter models is not currently supported - an internet "
                        "connection will be required during evaluation in this case. "
                        f"{open_issue_msg}",
                        level=logging.WARNING,
                    )

            loaded_model: "BenchmarkModule | None" = None
            for dataset_config in model_config_to_dataset_configs[model_config]:
                # Revert any changes to the benchmark configuration made for the
                # previous dataset
                for param, value in benchmark_params_to_revert.items():
                    setattr(benchmark_config, param, value)
                benchmark_params_to_revert = dict()

                # Update the benchmark config if the dataset requires it
                if (
                    "val" not in dataset_config.splits
                    and not benchmark_config.evaluate_test_split
                ):
                    log(
                        "The dataset does not have a validation split, so even though "
                        "you requested evaluating the validation split (the default), "
                        "we will evaluate on the test split.",
                        level=logging.DEBUG,
                    )
                    benchmark_params_to_revert["evaluate_test_split"] = False
                    benchmark_config.evaluate_test_split = True
                if dataset_config.task.requires_zero_shot and benchmark_config.few_shot:
                    log(
                        "The task requires zero-shot evaluation, so even though you "
                        "requested few-shot evaluation (the default), we will evaluate "
                        "zero-shot.",
                        level=logging.DEBUG,
                    )
                    benchmark_params_to_revert["few_shot"] = True
                    benchmark_config.few_shot = False

                # We do not re-initialise generative models as their architecture is not
                # customised to specific datasets
                if model_config.model_type == ModelType.GENERATIVE:
                    if loaded_model is None:
                        try:
                            loaded_model = load_model(
                                model_config=model_config,
                                dataset_config=dataset_config,
                                benchmark_config=benchmark_config,
                            )
                        except InvalidModel as e:
                            if benchmark_config.raise_errors:
                                raise e
                            log(e.message, level=logging.ERROR)

                            # Add the remaining number of benchmarks for the model to
                            # our benchmark counter, since we're skipping the rest of
                            # them
                            num_finished_benchmarks += (
                                len(dataset_configs)
                                - dataset_configs.index(dataset_config)
                                - 1
                            )
                            break
                    else:
                        loaded_model.dataset_config = dataset_config

                    # Skip the benchmark if the model is not of the correct
                    # generative type
                    if (
                        loaded_model.generative_type
                        not in dataset_config.allowed_generative_types
                    ):
                        log(
                            f"Skipping the benchmark of model "
                            f"{model_config.model_id!r}on dataset "
                            f"{dataset_config.name!r} because the model has generative "
                            f"type {loaded_model.generative_type} and the dataset "
                            f"only allows {dataset_config.allowed_generative_types}.",
                            level=logging.DEBUG,
                        )
                        num_finished_benchmarks += 1
                        continue

                # Benchmark a single model on a single dataset
                benchmark_output_or_err = self._benchmark_single(
                    model=loaded_model,
                    model_config=model_config,
                    dataset_config=dataset_config,
                    benchmark_config=benchmark_config,
                    num_finished_benchmarks=num_finished_benchmarks,
                    num_total_benchmarks=total_benchmarks,
                )

                if (
                    isinstance(benchmark_output_or_err, Exception)
                    and benchmark_config.raise_errors
                ):
                    raise benchmark_output_or_err

                elif isinstance(benchmark_output_or_err, InvalidBenchmark):
                    log(benchmark_output_or_err.message, level=logging.WARNING)
                    num_finished_benchmarks += 1
                    continue

                elif isinstance(benchmark_output_or_err, InvalidModel):
                    log(benchmark_output_or_err.message, level=logging.WARNING)

                    # Add the remaining number of benchmarks for the model to our
                    # benchmark counter, since we're skipping the rest of them
                    num_finished_benchmarks += (
                        len(dataset_configs) - dataset_configs.index(dataset_config) - 1
                    )
                    break

                else:
                    record: BenchmarkResult = benchmark_output_or_err
                    current_benchmark_results.append(record)
                    if benchmark_config.save_results:
                        record.append_to_results(results_path=self.results_path)

                num_finished_benchmarks += 1

            del loaded_model
            if benchmark_config.clear_model_cache:
                clear_model_cache_fn(cache_dir=benchmark_config.cache_dir)

        log(
            f"\nCompleted {num_finished_benchmarks:,} benchmarks.\n", level=logging.INFO
        )

        # This avoids the following warning at the end of the benchmarking:
        #   Warning: WARNING: process group has NOT been destroyed before we destruct
        #   ProcessGroupNCCL. On normal program exit, the application should call
        #   destroy_process_group to ensure that any pending NCCL operations have
        #   finished in this process. In rare cases this process can exit before this
        #   point and block the progress of another member of the process group. This
        #   constraint has always been present,  but this warning has only been added
        #   since PyTorch 2.4 (function operator())
        with contextlib.suppress(AssertionError):
            destroy_process_group()
        return current_benchmark_results

    def _prepare_model_ids(self, model_id: c.Sequence[str] | str) -> c.Sequence[str]:
        """Prepare the model ID(s) to be benchmarked.

        Args:
            model_id:
                The model ID(s) of the models to benchmark.

        Returns:
            The prepared list of model IDs.
        """
        model_ids = [model_id] if isinstance(model_id, str) else model_id

        # Reorder the `model_ids` list to include the ones present in the benchmark
        # results first
        benchmarked_model_ids = [
            re.sub(r"\(.+\)", "", record.model).strip()
            for record in self.benchmark_results
        ]
        model_ids_sorted = [m_id for m_id in model_ids if m_id in benchmarked_model_ids]
        model_ids_sorted += [
            m_id for m_id in model_ids if m_id not in benchmarked_model_ids
        ]

        return [m_id.rstrip(" /") for m_id in model_ids_sorted]

    def _benchmark_single(
        self,
        model: "BenchmarkModule | None",
        model_config: "ModelConfig",
        dataset_config: "DatasetConfig",
        benchmark_config: "BenchmarkConfig",
        num_finished_benchmarks: int,
        num_total_benchmarks: int,
    ) -> BenchmarkResult | InvalidBenchmark | InvalidModel:
        """Benchmark a single model on a single dataset.

        Args:
            model:
                The model to benchmark.
            model_config:
                The configuration of the model we are evaluating.
            dataset_config:
                The configuration of the dataset we are evaluating on.
            benchmark_config:
                The general benchmark configuration.
            num_finished_benchmarks:
                The number of benchmarks that have already been completed.
            num_total_benchmarks:
                The total number of benchmarks to be completed.

        Returns:
            The benchmark result, or an error if the benchmark was unsuccessful.

        Raises:
            RuntimeError:
                If the MPS fallback is not enabled when required.
            InvalidBenchmark:
                If the benchmark was unsuccessful.
            InvalidModel:
                If the model is invalid.
        """
        for _ in range(num_attempts := 5):
            try:
                # Set random seeds to enforce reproducibility of the randomly
                # initialised weights
                rng = enforce_reproducibility()

                if model is None or model_config.model_type != ModelType.GENERATIVE:
                    model = load_model(
                        model_config=model_config,
                        dataset_config=dataset_config,
                        benchmark_config=benchmark_config,
                    )
                assert model is not None

                initial_logging(
                    model_config=model_config,
                    dataset_config=dataset_config,
                    benchmark_config=benchmark_config,
                    num_finished_benchmarks=num_finished_benchmarks,
                    num_total_benchmarks=num_total_benchmarks,
                )

                if dataset_config.task == SPEED:
                    scores = benchmark_speed(
                        model=model, benchmark_config=benchmark_config
                    )

                else:
                    bootstrapped_datasets = load_data(
                        rng=rng,
                        dataset_config=dataset_config,
                        benchmark_config=benchmark_config,
                    )
                    prepared_datasets = model.prepare_datasets(
                        datasets=bootstrapped_datasets, task=dataset_config.task
                    )
                    if model_config.model_type == ModelType.GENERATIVE:
                        scores = generate(
                            model=model,
                            datasets=prepared_datasets,
                            model_config=model_config,
                            dataset_config=dataset_config,
                            benchmark_config=benchmark_config,
                        )
                    else:
                        scores = finetune(
                            model=model,
                            datasets=prepared_datasets,
                            model_config=model_config,
                            dataset_config=dataset_config,
                            benchmark_config=benchmark_config,
                        )

                results = log_scores(
                    dataset_name=dataset_config.logging_string,
                    metrics=dataset_config.task.metrics,
                    scores=scores,
                    model_id=model_config.model_id,
                    model_revision=model_config.revision,
                    model_param=model_config.param,
                )

                model_id_to_be_stored = model_config.model_id
                if model_config.revision != "main":
                    model_id_to_be_stored += f"@{model_config.revision}"
                if model_config.param is not None:
                    model_id_to_be_stored += f"#{model_config.param}"

                record = BenchmarkResult(
                    dataset=dataset_config.name,
                    task=dataset_config.task.name,
                    languages=[language.code for language in dataset_config.languages],
                    model=model_id_to_be_stored,
                    results=results,
                    num_model_parameters=model.num_params,
                    max_sequence_length=model.model_max_length,
                    vocabulary_size=model.vocab_size,
                    merge=model_config.merge,
                    generative=model_config.model_type == ModelType.GENERATIVE,
                    generative_type=(
                        model.generative_type.value
                        if model.generative_type is not None
                        else None
                    ),
                    few_shot=(
                        None
                        if dataset_config.task.requires_zero_shot
                        else benchmark_config.few_shot
                    ),
                    validation_split=(
                        None
                        if "val" not in dataset_config.splits
                        else not benchmark_config.evaluate_test_split
                    ),
                )
                log(f"Results:\n{results}", level=logging.DEBUG)
                return record

            except HuggingFaceHubDown:
                wait_time = 30
                log(
                    f"The Hugging Face Hub seems to be down. Retrying in {wait_time} "
                    "seconds.",
                    level=logging.DEBUG,
                )
                sleep(wait_time)
                continue

            except (InvalidBenchmark, InvalidModel) as e:
                # If the model ID is not valid then raise an error
                model_err_msg = "does not exist on the Hugging Face Hub"
                if benchmark_config.raise_errors and model_err_msg in str(e):
                    raise e

                # Otherwise, if the error is due to the MPS fallback not being enabled,
                # then raise an error asking the user to enable it
                elif "PYTORCH_ENABLE_MPS_FALLBACK" in str(e):
                    raise RuntimeError(
                        "The benchmark failed because the environment variable "
                        "`PYTORCH_ENABLE_MPS_FALLBACK` is not set. Please set this "
                        "environment variable to `1` and try again."
                    )

                elif benchmark_config.raise_errors:
                    raise e
                return e
        else:
            return InvalidBenchmark(
                f"Failed to benchmark model {model_config.model_id!r} on dataset "
                f"{dataset_config.name!r} after {num_attempts} attempts."
            )

    def __call__(self, *args: t.Any, **kwds: t.Any) -> t.Any:  # noqa: ANN401
        """Alias for `self.benchmark()`."""
        log(
            "Calling the `Benchmarker` class directly is deprecated. Please use the "
            "`benchmark` function instead. This will be removed in a future version.",
            level=logging.WARNING,
        )
        return self.benchmark(*args, **kwds)


def get_record(
    model_config: "ModelConfig",
    dataset_config: "DatasetConfig",
    benchmark_config: "BenchmarkConfig",
    benchmark_results: c.Sequence[BenchmarkResult],
) -> BenchmarkResult | None:
    """Get the benchmark record for a given model and dataset.

    Args:
        model_config:
            The configuration of the model we are evaluating.
        dataset_config:
            The configuration of the dataset we are evaluating on.
        benchmark_config:
            The general benchmark configuration.
        benchmark_results:
            The benchmark results.

    Returns:
        The benchmark record, or None if no such record exists.
    """
    for record in benchmark_results:
        model_id_components = split_model_id(model_id=record.model)
        same_model_id = model_id_components.model_id == model_config.model_id
        same_revision = model_id_components.revision == model_config.revision
        same_param = model_id_components.param == model_config.param
        same_dataset = record.dataset == dataset_config.name
        same_split = record.validation_split != benchmark_config.evaluate_test_split
        same_num_shots = (
            record.few_shot == benchmark_config.few_shot
            or record.few_shot is None
            or not record.generative
            or dataset_config.task.requires_zero_shot
        )
        if (
            same_model_id
            and same_revision
            and same_param
            and same_dataset
            and same_split
            and same_num_shots
        ):
            return record
    return None


def clear_model_cache_fn(cache_dir: str) -> None:
    """Clear the model cache.

    Note that this will not remove the stored completions.

    Args:
        cache_dir:
            The path to the cache directory.
    """
    model_cache_path = Path(cache_dir) / "model_cache"
    model_cache_path.mkdir(parents=True, exist_ok=True)
    for model_dir in model_cache_path.iterdir():
        if model_dir.is_dir():
            for sub_model_dir in model_dir.iterdir():
                if sub_model_dir.is_dir():
                    rmtree(sub_model_dir)


def prepare_dataset_configs(
    dataset_names: c.Sequence[str], custom_datasets_file: Path
) -> c.Sequence["DatasetConfig"]:
    """Prepare the dataset configuration(s) to be benchmarked.

    Args:
        dataset_names:
            The dataset names to benchmark.
        custom_datasets_file:
            A path to a Python file containing custom dataset configurations.

    Returns:
        The prepared list of model IDs.
    """
    return [
        cfg
        for cfg in get_all_dataset_configs(
            custom_datasets_file=custom_datasets_file
        ).values()
        if cfg.name in dataset_names
    ]


def initial_logging(
    model_config: "ModelConfig",
    dataset_config: "DatasetConfig",
    benchmark_config: "BenchmarkConfig",
    num_finished_benchmarks: int,
    num_total_benchmarks: int,
) -> None:
    """Initial logging at the start of the benchmarking process.

    Args:
        model_config:
            The configuration of the model we are evaluating.
        dataset_config:
            The configuration of the dataset we are evaluating on.
        benchmark_config:
            The general benchmark configuration.
        num_finished_benchmarks:
            The number of benchmarks that have already been finished.
        num_total_benchmarks:
            The total number of benchmarks to be run.
    """
    model_id = model_config.model_id
    if model_config.revision and model_config.revision != "main":
        model_id += f"@{model_config.revision}"
    if model_config.param is not None:
        model_id += f"#{model_config.param}"

    split_type = "validation" if not benchmark_config.evaluate_test_split else "test"
    if model_config.task in GENERATIVE_PIPELINE_TAGS:
        if benchmark_config.few_shot:
            eval_type = "Few-shot benchmarking"
        else:
            eval_type = "Zero-shot benchmarking"
    else:
        eval_type = "Benchmarking"

    log_once(
        f"\n{eval_type} {model_id} on the {split_type} split of "
        f"{dataset_config.logging_string} ({num_finished_benchmarks + 1}/"
        f"{num_total_benchmarks} benchmarks)...",
        prefix=f"\n[{dt.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}]",
        level=logging.INFO,
    )

    if dataset_config.unofficial:
        log_once(
            f"Note that the {dataset_config.name!r} dataset is unofficial, "
            "meaning that the resulting evaluation will not be included in the "
            "official leaderboard.",
            level=logging.WARNING,
        )

    if benchmark_config.debug:
        log_once(
            "Running in debug mode. This will output additional information, as "
            "well as store the model outputs in the current directory after each "
            "batch. For this reason, evaluation will be slower.",
            level=logging.WARNING,
        )



"""ModelCache class for caching model outputs."""

import collections.abc as c
import hashlib
import json
import logging
import sys
import typing as t
from collections import defaultdict
from dataclasses import asdict

from datasets import Dataset

from .constants import NUM_GENERATION_TOKENS_FOR_CLASSIFICATION
from .data_models import GenerativeModelOutput, SingleGenerativeModelOutput
from .logging_utils import get_pbar, log, log_once

if t.TYPE_CHECKING:
    from pathlib import Path


class ModelCache:
    """A cache for model outputs.

    Attributes:
        model_cache_dir:
            The directory to store the cache in.
        cache_path:
            The path to the cache file.
        cache:
            The model output cache.
        max_generated_tokens:
            The maximum number of tokens to generate for each example.
        progress_bar:
            Whether to show a progress bar when caching model outputs.
        hash_inputs:
            Whether to hash the model inputs to use as keys in the cache.
    """

    def __init__(
        self,
        model_cache_dir: "Path",
        cache_name: str,
        max_generated_tokens: int,
        progress_bar: bool,
        hash_inputs: bool,
    ) -> None:
        """Initialise the model output cache.

        Args:
            model_cache_dir:
                The directory to store the cache in.
            cache_name:
                The name of the cache file.
            max_generated_tokens:
                The maximum number of tokens to generate for each example.
            progress_bar:
                Whether to show a progress bar when caching model outputs.
            hash_inputs:
                Whether to hash the model inputs to use as keys in the cache.
        """
        self.model_cache_dir = model_cache_dir
        self.model_cache_dir.mkdir(parents=True, exist_ok=True)
        self.cache_path = self.model_cache_dir / cache_name.replace("/", "--")
        self.max_generated_tokens = max_generated_tokens
        self.progress_bar = progress_bar
        self.hash_inputs = hash_inputs

    def load(self) -> None:
        """Load the model output cache."""
        if not self.cache_path.exists():
            with self.cache_path.open("w") as f:
                json.dump(dict(), f)

        try:
            with self.cache_path.open() as f:
                json_cache = json.load(f)
        except json.JSONDecodeError:
            log(
                f"Failed to load the cache from {self.cache_path}. The cache will be "
                f"re-initialised.",
                level=logging.WARNING,
            )
            json_cache = dict()
            with self.cache_path.open("w") as f:
                json.dump(dict(), f)

        cache: dict[str, SingleGenerativeModelOutput] = dict()
        for key in json_cache:
            cache[key] = SingleGenerativeModelOutput(**json_cache[key])

        self.cache = cache

    def save(self) -> None:
        """Save the model output cache to disk."""
        dumpable_cache: dict[str, dict] = defaultdict(dict)
        for key, value in self.cache.items():
            dumpable_cache[key] = asdict(value)

        try:
            with self.cache_path.open("w") as f:
                json.dump(dumpable_cache, f)
        except KeyError:
            log(
                f"Failed to load the cache from {self.cache_path}. The cache will be "
                f"re-initialised.",
                level=logging.WARNING,
            )
            self.cache = dict()
            with self.cache_path.open("w") as f:
                json.dump(dict(), f)

    def _hash_key(self, key: str | c.Sequence[dict[str, str]]) -> str:
        """Hash the key to use as an index in the cache.

        Args:
            key:
                The key to hash.

        Returns:
            The hashed key.
        """
        if self.hash_inputs:
            return hashlib.md5(string=str(key).encode()).hexdigest()
        else:
            return str(key)

    def __getitem__(
        self, key: str | c.Sequence[dict[str, str]]
    ) -> SingleGenerativeModelOutput:
        """Get an item from the cache.

        Args:
            key:
                The key to use to index the cache.

        Returns:
            The model output.
        """
        hashed_key = self._hash_key(key=key)
        return self.cache[hashed_key]

    def __setitem__(
        self, key: str | c.Sequence[dict[str, str]], value: SingleGenerativeModelOutput
    ) -> None:
        """Set an item in the cache.

        Args:
            key:
                The key to use to index the cache.
            value:
                The value to set in the cache.
        """
        hashed_key = self._hash_key(key=key)
        self.cache[hashed_key] = value

    def remove(self) -> None:
        """Remove the cache from memory and delete it from disk."""
        self.cache_path.unlink()
        del self.cache

    def __contains__(self, key: str | c.Sequence[dict[str, str]]) -> bool:
        """Check if a key is in the cache.

        Args:
            key:
                The key to check.

        Returns:
            Whether the key is in the cache.
        """
        hashed_key = self._hash_key(key=key)
        return hashed_key in self.cache

    def add_to_cache(
        self, model_inputs: dict, model_output: GenerativeModelOutput
    ) -> None:
        """Add the model input/output to the cache.

        Args:
            model_inputs:
                The model inputs.
            model_output:
                The model output.
        """
        input_column = "messages" if "messages" in model_inputs else "text"
        model_inputs = model_inputs[input_column]

        # Double check that the number of inputs and outputs match
        if not len(model_inputs) == len(model_output.sequences):
            log(
                f"Number of model inputs ({len(model_inputs)}) does not match the "
                f"number of model outputs ({len(model_output.sequences)}). We will not "
                f"cache the model outputs.",
                level=logging.WARNING,
            )
            return

        # Store the generated sequences in the cache, one by one
        with get_pbar(
            iterable=model_inputs,
            desc="Caching model outputs",
            disable=hasattr(sys, "_called_from_test") or not self.progress_bar,
        ) as pbar:
            for sample_idx, model_input in enumerate(pbar):
                # Extract the scores from the model output, to be cached. We only store
                # the indices of the top scores, to save space. Further, we only store
                # the scores if the generated sequence is shorter than the maximum
                # length
                if (
                    model_output.scores is not None
                    and self.max_generated_tokens
                    <= NUM_GENERATION_TOKENS_FOR_CLASSIFICATION
                ):
                    assert model_output.scores is not None
                    scores = model_output.scores[sample_idx]
                else:
                    if model_output.scores is not None:
                        log_once(
                            "The generated sequence is longer than the maximum "
                            "length for classification. Not caching the scores.",
                            level=logging.DEBUG,
                        )
                    scores = None
                self[model_input] = SingleGenerativeModelOutput(
                    sequence=model_output.sequences[sample_idx], scores=scores
                )


def split_dataset_into_cached_and_non_cached(
    dataset: "Dataset", cache: ModelCache
) -> tuple["Dataset", "Dataset"]:
    """Split a dataset into a cached and non-cached part.

    Args:
        dataset:
            The dataset to split.
        cache:
            The model output cache.

    Returns:
        The cached and non-cached parts of the dataset.
    """
    # Get the sample indices of the non-cached examples, which are unique with respect
    # to the "text" column.
    input_column = "messages" if "messages" in dataset.column_names else "text"
    dataset_texts = dataset[input_column]
    unique_non_cached_ids = set()
    unique_texts = list()
    for idx, dataset_text in enumerate(dataset_texts):
        if dataset_text not in cache and dataset_text not in unique_texts:
            unique_non_cached_ids.add(idx)
            unique_texts.append(dataset_text)

    # The cached examples are the ones that are not in the non-cached examples. This
    # means that if the dataset has duplicates, only a single copy of the duplicate
    # will be put in the non-cached part, and the rest in the cached part.
    cached_ids = set(range(len(dataset))) - unique_non_cached_ids

    cached = dataset.select(cached_ids)
    non_cached = dataset.select(unique_non_cached_ids)

    assert isinstance(cached, Dataset), (
        f"Expected the cached dataset to be a Dataset, but got {type(cached)}"
    )
    assert isinstance(non_cached, Dataset), (
        f"Expected the non-cached dataset to be a Dataset, but got {type(non_cached)}"
    )
    return cached, non_cached


def load_cached_model_outputs(
    cached_dataset: "Dataset", cache: ModelCache
) -> GenerativeModelOutput:
    """Load the cached model outputs.

    Args:
        cached_dataset:
            The dataset containing the cached examples.
        cache:
            The model output cache.

    Returns:
        The model output containing the cached sequences.
    """
    input_column = "messages" if "messages" in cached_dataset.column_names else "text"
    cached_model_outputs: c.Sequence[SingleGenerativeModelOutput] = [
        cache[prompt] for prompt in cached_dataset[input_column]
    ]

    cached_sequences = [model_output.sequence for model_output in cached_model_outputs]

    if cached_model_outputs[0].scores is None:
        return GenerativeModelOutput(sequences=cached_sequences)

    cached_scores = [model_output.scores or [] for model_output in cached_model_outputs]
    return GenerativeModelOutput(sequences=cached_sequences, scores=cached_scores)



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==2.19.1",
#     "nltk==3.8.1",
#     "huggingface-hub==0.24.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the Público-mini summarisation dataset."""

import random

import nltk
from datasets import Dataset, DatasetDict, load_dataset
from huggingface_hub.hf_api import HfApi
from nltk.tokenize import sent_tokenize  # noqa: E402

nltk.download("punkt")

TOTAL = 1024 + 256 + 2048


def main() -> None:
    """Create the Público-mini dataset."""
    raw = load_dataset("duarteocarmo/cc_news_publico", split="train")
    assert isinstance(raw, Dataset)
    processed = [_extract_fields(example=dict(x)) for x in raw]
    processed = [x for x in processed if x]

    random.seed(42)
    random.shuffle(processed)
    processed = processed[:TOTAL]

    train = Dataset.from_list(processed[:1024])
    val = Dataset.from_list(processed[1024 : 1024 + 256])
    test = Dataset.from_list(processed[1024 + 256 :])

    dataset = DatasetDict({"train": train, "val": val, "test": test})

    # Push the dataset to the Hugging Face Hub
    dataset_id = "EuroEval/publico-mini"
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)
    dataset.push_to_hub(dataset_id, private=True)


def _extract_fields(example: dict) -> dict | None:
    """Extracts fields from a dataset example.

    Args:
        example: Dictionary with keys 'title' and 'plain_text'.

    Returns:
        dict with 'text' and 'target_text', or None if invalid.
    """
    title = example.get("title", "").strip()
    text = example.get("plain_text", "").strip()
    if not title or not text:
        return None
    sentences = sent_tokenize(text, language="portuguese")
    if len(sentences) < 3:
        return None
    return {
        "text": f"{title}\n\n" + " ".join(sentences[2:]),
        "target_text": " ".join(sentences[:2]),
    }


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the PONER NER dataset and upload it to the HF Hub."""

import pandas as pd
from datasets import Dataset, DatasetDict, Split, load_dataset
from huggingface_hub import HfApi


def main() -> None:
    """Create the PONER NER dataset and upload it to the HF Hub."""
    # Define dataset ID
    repo_id = "romanjanik/PONER"

    # Load the dataset
    dataset = load_dataset(path=repo_id, token=True)
    assert isinstance(dataset, DatasetDict)

    # Convert the dataset to dataframes
    train_df = dataset["train"].to_pandas()
    test_df = dataset["test"].to_pandas()
    val_df = dataset["dev"].to_pandas()

    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(test_df, pd.DataFrame)
    assert isinstance(val_df, pd.DataFrame)

    # Process the dataframes
    columns_to_drop = [
        col for col in train_df.columns if col not in ["tokens", "ner_tags"]
    ]
    train_df.drop(columns=columns_to_drop, inplace=True)
    test_df.drop(columns=columns_to_drop, inplace=True)
    val_df.drop(columns=columns_to_drop, inplace=True)

    train_df["text"] = train_df["tokens"].map(lambda tokens: " ".join(tokens))
    test_df["text"] = test_df["tokens"].map(lambda tokens: " ".join(tokens))
    val_df["text"] = val_df["tokens"].map(lambda tokens: " ".join(tokens))

    train_df.rename(columns={"ner_tags": "labels"}, inplace=True)
    test_df.rename(columns={"ner_tags": "labels"}, inplace=True)
    val_df.rename(columns={"ner_tags": "labels"}, inplace=True)

    # Create label mapping
    label_names = dataset["train"].features["ner_tags"].feature.names
    label_mapping = create_label_mapping(label_names)

    # Convert the NER tags from IDs to strings using our mapping
    train_df["labels"] = train_df["labels"].map(
        lambda ner_tags: [label_mapping.get(ner_tag, "O") for ner_tag in ner_tags]
    )
    test_df["labels"] = test_df["labels"].map(
        lambda ner_tags: [label_mapping.get(ner_tag, "O") for ner_tag in ner_tags]
    )
    val_df["labels"] = val_df["labels"].map(
        lambda ner_tags: [label_mapping.get(ner_tag, "O") for ner_tag in ner_tags]
    )

    # Create splits
    train_size = 1024
    val_size = 256
    test_size = 2048

    train_df = train_df.sample(n=min(train_size, len(train_df)), random_state=4242)
    val_df = val_df.sample(n=min(val_size, len(val_df)), random_state=4242)
    test_df = test_df.sample(n=min(test_size, len(test_df)), random_state=4242)

    # Reset the index
    train_df = train_df.reset_index(drop=True)
    val_df = val_df.reset_index(drop=True)
    test_df = test_df.reset_index(drop=True)

    # Collect datasets in a dataset dictionary
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(test_df, split=Split.TEST),
        }
    )

    # Create dataset ID
    dataset_id = "EuroEval/poner-mini"

    # Remove the dataset from Hugging Face Hub if it already exists
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)

    # Push the dataset to the Hugging Face Hub
    dataset.push_to_hub(dataset_id, private=True)


def create_label_mapping(label_names: list[str]) -> dict[int, str]:
    """Create mapping from PONER labels to standard BIO labels.

    Args:
        label_names: The list of label names.

    Returns:
        The mapping from PONER labels to standard BIO labels.
    """
    # Define a base mapping for the label prefixes
    base_mapping = {
        "B-p": "B-PER",
        "I-p": "I-PER",
        "B-i": "B-ORG",
        "I-i": "I-ORG",
        "B-g": "B-LOC",
        "I-g": "I-LOC",
        "B-t": "B-MISC",
        "I-t": "I-MISC",
        "B-o": "B-MISC",
        "I-o": "I-MISC",
    }

    # Create the mapping using the base mapping
    mapping = {
        i: base_mapping.get(label_name, "O") for i, label_name in enumerate(label_names)
    }

    return mapping


if __name__ == "__main__":
    main()



"""Functions related to text generation of models."""

import collections.abc as c
import logging
import sys
import typing as t
from pathlib import Path

from datasets import Dataset
from tqdm.auto import tqdm

from .enums import BatchingPreference, TaskGroup
from .exceptions import InvalidBenchmark, InvalidModel
from .logging_utils import get_pbar, log, log_once
from .model_cache import (
    ModelCache,
    load_cached_model_outputs,
    split_dataset_into_cached_and_non_cached,
)
from .utils import clear_memory

if t.TYPE_CHECKING:
    from datasets import DatasetDict

    from .benchmark_modules import BenchmarkModule
    from .data_models import (
        BenchmarkConfig,
        DatasetConfig,
        GenerativeModelOutput,
        ModelConfig,
    )


def generate(
    model: "BenchmarkModule",
    datasets: c.Sequence["DatasetDict"],
    model_config: "ModelConfig",
    dataset_config: "DatasetConfig",
    benchmark_config: "BenchmarkConfig",
) -> c.Sequence[dict[str, float]]:
    """Evaluate a model on a dataset through generation.

    Args:
        model:
            The model to evaluate.
        datasets:
            The datasets to evaluate on.
        model_config:
            The configuration of the model.
        benchmark_config:
            The configuration of the benchmark.
        dataset_config:
            The configuration of the dataset.

    Returns:
        A list of dictionaries containing the test scores.
    """
    # Set up the name of the model output cache. If we are testing then we save the
    # model outputs to a different cache and ensure that that cache is deleted before
    # the next test, to ensure that the tests are independent of each other
    if benchmark_config.debug:
        model_cache_dir = Path.cwd()
    else:
        model_cache_dir = Path(model_config.model_cache_dir)
    if hasattr(sys, "_called_from_test"):
        cache_name = f"{dataset_config.name}-model-outputs-test.json"
        (model_cache_dir / cache_name).unlink(missing_ok=True)
    elif benchmark_config.debug:
        cache_name = f"{model_config.model_id}-{dataset_config.name}-model-outputs.json"
    else:
        cache_name = f"{dataset_config.name}-model-outputs.json"

    cache = ModelCache(
        model_cache_dir=model_cache_dir,
        cache_name=cache_name,
        max_generated_tokens=dataset_config.max_generated_tokens,
        progress_bar=benchmark_config.progress_bar,
        hash_inputs=not benchmark_config.debug,
    )

    scores: list[dict[str, float]] = list()
    for idx in get_pbar(
        iterable=range(len(datasets)),
        desc="Benchmarking",
        disable=not benchmark_config.progress_bar,
    ):
        test_scores = generate_single_iteration(
            model=model,
            dataset=datasets[idx]["test"],
            cache=cache,
            dataset_config=dataset_config,
            benchmark_config=benchmark_config,
        )
        log(f"Test scores for iteration {idx}: {test_scores}", level=logging.DEBUG)
        scores.append(test_scores)
        clear_memory()

    if not benchmark_config.debug:
        cache.remove()

    return scores


def generate_single_iteration(
    dataset: "Dataset",
    model: "BenchmarkModule",
    dataset_config: "DatasetConfig",
    benchmark_config: "BenchmarkConfig",
    cache: ModelCache,
) -> dict[str, float]:
    """Evaluate a model on a dataset in a single iteration through generation.

    Args:
        dataset:
            The dataset to evaluate on.
        model:
            The model to evaluate.
        dataset_config:
            The configuration of the dataset.
        benchmark_config:
            The configuration of the benchmark.
        cache:
            The model output cache.

    Returns:
        A list of dictionaries containing the scores for each metric.
    """
    cache.load()

    # Split up the dataset into a cached and non-cached part, unless we are not
    # bootstrapping the samples. In that case, we just use the dataset as is.
    if dataset_config.bootstrap_samples:
        cached_dataset, non_cached_dataset = split_dataset_into_cached_and_non_cached(
            dataset=dataset, cache=cache
        )
    else:
        cached_dataset = Dataset.from_dict({})
        non_cached_dataset = dataset

    all_preds: list[str] = list()

    if len(non_cached_dataset) > 0:
        itr: t.Iterable
        match model.batching_preference:
            case BatchingPreference.SINGLE_SAMPLE:
                itr = get_pbar(
                    iterable=non_cached_dataset,
                    disable=not benchmark_config.progress_bar,
                )
            case BatchingPreference.ALL_AT_ONCE:
                itr = [non_cached_dataset[:]]
            case _:
                raise InvalidModel(
                    f"The batching preference {model.batching_preference!r} is "
                    "currently not supported."
                )
                # NOTE: The code below can be used if we want to support batching for
                # generative models. But in that case, we have to deal with the naming
                # of the batch size variable, since it is currently
                # `finetuning_batch_size`, as it is only used during finetuning of
                # encoder models.
                # num_batches = len(non_cached_dataset) // benchmark_config.batch_size
                # if len(non_cached_dataset) % benchmark_config.batch_size != 0:
                #     num_batches += 1
                # itr = get_pbar(
                #     iterable=mit.batched(
                #         iterable=non_cached_dataset, n=benchmark_config.batch_size
                #     ),
                #     total=len(non_cached_dataset) // benchmark_config.batch_size,
                # )

        # Generate the completions for the non-cached examples
        for batch in itr:
            assert isinstance(batch, dict)

            single_sample_batch = (
                "text" in batch and isinstance(batch["text"], str)
            ) or ("messages" in batch and isinstance(batch["messages"][0], dict))
            if single_sample_batch:
                batch = {key: [value] for key, value in batch.items()}

            model_output = model.generate(inputs=batch)
            extracted_labels = model.extract_labels_from_generation(
                input_batch=batch, model_output=model_output
            )

            # Extended logging if we are running in debug mode
            if benchmark_config.debug:
                debug_log(
                    batch=batch,
                    model_output=model_output,
                    extracted_labels=extracted_labels,  # type: ignore[arg-type]
                    dataset_config=dataset_config,
                )

            cache.add_to_cache(model_inputs=batch, model_output=model_output)
            all_preds.extend(extracted_labels)

            # If we are debugging then we save the cache often, but since this makes
            # evaluation slower, we do not do this by default
            if benchmark_config.debug:
                cache.save()

        if isinstance(itr, tqdm):
            itr.close()

        # Store the cache to disk
        cache.save()

    # Fetch the cached predictions for the cached examples
    if len(cached_dataset) > 0:
        model_output = load_cached_model_outputs(
            cached_dataset=cached_dataset, cache=cache
        )
        extracted_labels = model.extract_labels_from_generation(
            input_batch=cached_dataset[:], model_output=model_output
        )
        all_preds.extend(extracted_labels)

    if "label" in non_cached_dataset.column_names:
        non_cached_labels = non_cached_dataset["label"]
        if not isinstance(non_cached_labels, list):
            non_cached_labels = list(non_cached_labels)
        cached_labels = cached_dataset["label"]
        if not isinstance(cached_labels, list):
            cached_labels = list(cached_labels)
        ground_truth = [
            label.lower() if isinstance(label, str) else label
            for label in non_cached_labels + cached_labels
        ]
    elif "labels" in non_cached_dataset.column_names:
        non_cached_labels = non_cached_dataset["labels"]
        if not isinstance(non_cached_labels, list):
            non_cached_labels = list(non_cached_labels)
        cached_labels = cached_dataset["labels"]
        if not isinstance(cached_labels, list):
            cached_labels = list(cached_labels)
        ground_truth = [
            [label.lower() if isinstance(label, str) else label for label in label_list]
            for label_list in non_cached_labels + cached_labels
        ]
    elif "target_text" in non_cached_dataset.column_names:
        non_cached_labels = non_cached_dataset["target_text"]
        if not isinstance(non_cached_labels, list):
            non_cached_labels = list(non_cached_labels)
        cached_labels = cached_dataset["target_text"]
        if not isinstance(cached_labels, list):
            cached_labels = list(cached_labels)
        ground_truth = non_cached_labels + cached_labels
    else:
        log_once(
            "No labels found in the dataset. We assume that this is intentional, and "
            "will not supply any ground truth labels for evaluation.",
            level=logging.DEBUG,
        )
        ground_truth = []

    itr_scores: dict[str, float] = model.compute_metrics(
        model_outputs_and_labels=(all_preds, ground_truth),
        dataset=dataset,
        benchmark_config=benchmark_config,
    )

    return itr_scores


def debug_log(
    batch: dict[str, t.Any],
    model_output: "GenerativeModelOutput",
    extracted_labels: c.Sequence[dict | str | c.Sequence[str]],
    dataset_config: "DatasetConfig",
) -> None:
    """Log inputs and outputs for debugging purposes.

    Args:
        batch:
            The batch of examples to evaluate on.
        model_output:
            The output of the model.
        extracted_labels:
            The extracted labels from the model output.
        dataset_config:
            The configuration of the dataset.
    """
    match dataset_config.task.task_group:
        case TaskGroup.TOKEN_CLASSIFICATION:
            log_msgs = [""]
            for tokens, predictions, labels in zip(
                batch["tokens"], extracted_labels, batch["labels"]
            ):
                predictions = [tag.upper() for tag in predictions]
                sample = list(zip(tokens, predictions, labels))
                log_batches = [
                    [("Tokens: ", "Predictions: ", "Labels: ")] + sample[i : i + 10]
                    for i in range(0, len(sample), 10)
                ]
                for log_batch in log_batches:
                    lengths = [len(max(triple, key=len)) for triple in log_batch]
                    log_batch = [
                        [f"{x:<{length}}" for x in triple]
                        for triple, length in zip(log_batch, lengths)
                    ]
                    tokens = [triple[0] for triple in log_batch]
                    predictions = [triple[1] for triple in log_batch]
                    labels = [triple[2] for triple in log_batch]
                    log_msgs.append(
                        "\t".join(tokens)
                        + "\n"
                        + "\t".join(predictions)
                        + "\n"
                        + "\t".join(labels)
                    )
            log("\n\n".join(log_msgs), level=logging.DEBUG)
            return

        case (
            TaskGroup.SEQUENCE_CLASSIFICATION | TaskGroup.MULTIPLE_CHOICE_CLASSIFICATION
        ):
            if "label" in batch:
                labels = [
                    dataset_config.prompt_label_mapping.get(label, label).lower()
                    for label in batch["label"]
                ]
            else:
                labels = [None] * len(extracted_labels)

        case TaskGroup.QUESTION_ANSWERING:
            extracted_labels = [
                prediction["prediction_text"]
                for prediction in extracted_labels
                if isinstance(prediction, dict)
            ]
            labels = [label["answers"]["text"][0] for label in batch["label"]]

        case TaskGroup.TEXT_TO_TEXT:
            labels = batch["target_text"]

        case _:
            raise InvalidBenchmark(
                f"The task group '{dataset_config.task.task_group}' is not supported."
            )

    if "messages" in batch:
        input_texts = [messages[-1]["content"] for messages in batch["messages"]]
    else:
        input_texts = batch["text"]

    metadata_keys: c.Sequence[str] = [
        key
        for key in batch.keys()
        if key not in ["text", "messages", "label", "labels", "target_text"]
    ]

    for idx in range(len(input_texts)):
        data_to_log: dict[str, t.Any] = {
            "Input": input_texts[idx],
            "Raw output": model_output.sequences[idx],
            "Prediction": extracted_labels[idx],
        }
        if labels[idx]:
            data_to_log["Label"] = labels[idx]
        data_to_log |= {key.capitalize(): batch[key][idx] for key in metadata_keys}
        log(
            "\n".join(f"{key}: {value!r}" for key, value in data_to_log.items()),
            level=logging.DEBUG,
        )



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the UNER-sk NER dataset and upload it to the HF Hub."""

import json
import re

import numpy as np
import pandas as pd
from datasets import Dataset, DatasetDict, load_dataset
from huggingface_hub import HfApi


def main() -> None:
    """Create the UNER-Slovak NER dataset and uploads it to the HF Hub."""
    repo_id = "universalner/uner_llm_inst_slovak"
    dataset = load_dataset(repo_id)

    # Assuming `inputs` is the text and `targets` contains the entities
    train_df = pd.DataFrame(dataset["train"])
    val_df = pd.DataFrame(dataset["validation"])
    test_df = pd.DataFrame(dataset["test"])

    train_df = process_df(train_df)
    val_df = process_df(val_df)
    test_df = process_df(test_df)

    # Make splits
    train_size = 1024
    val_size = 256
    test_size = 2048

    # Take additional samples from the training set for the test split
    additional_test_samples_needed = test_size - len(test_df)
    additional_test_samples = train_df.sample(
        n=additional_test_samples_needed, random_state=4242
    )
    final_test_df = pd.concat([test_df, additional_test_samples], ignore_index=True)
    remaining_train_df = train_df[~train_df.index.isin(additional_test_samples.index)]

    # Sample final splits
    final_test_df = final_test_df.sample(n=test_size, random_state=4242).reset_index(
        drop=True
    )
    final_val_df = val_df.sample(n=val_size, random_state=4242).reset_index(drop=True)
    final_train_df = remaining_train_df.sample(
        n=train_size, random_state=4242
    ).reset_index(drop=True)

    assert isinstance(final_train_df, pd.DataFrame)
    assert isinstance(final_val_df, pd.DataFrame)
    assert isinstance(final_test_df, pd.DataFrame)

    dataset_dict = DatasetDict(
        {
            "train": Dataset.from_pandas(final_train_df),
            "val": Dataset.from_pandas(final_val_df),
            "test": Dataset.from_pandas(final_test_df),
        }
    )

    dataset_id = "EuroEval/uner-sk-mini"
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)
    dataset_dict.push_to_hub(dataset_id, private=True)


def process_df(df: pd.DataFrame) -> pd.DataFrame:
    """Process the DataFrame.

    Args:
        df: The DataFrame to process.

    Returns:
        The processed DataFrame.
    """
    df = df.drop_duplicates()

    df["text"] = df["inputs"].apply(lambda x: x.split("nižšie?")[-1].strip())
    df["tokens"] = df["text"].apply(lambda x: _split_with_punctuation(x))
    df["positions"] = df["inputs"].apply(
        lambda x: _get_positions(x.split("nižšie?")[-1].strip())
    )
    df["labels"] = df.apply(
        lambda row: _assign_labels(
            row["tokens"], _parse_targets(row["targets"]), row["positions"], row["text"]
        ),
        axis=1,
    )
    keep_columns = ["tokens", "labels"]
    df = df.loc[keep_columns]
    return df


def _split_with_punctuation(text: str) -> list[str]:
    """Tokenize text, splitting on spaces and treating punctuation as separate tokens.

    Args:
        text: The text to tokenize.

    Returns:
        The tokenized text.
    """
    return re.findall(r"\w+|[^\w\s]", text)


def _parse_targets(target_str: str) -> list[dict]:
    """Parse the targets string into a list of dictionaries.

    Args:
        target_str: The targets string to parse.

    Returns:
        The parsed targets.
    """
    targets_json = json.loads(f"{{{target_str}}}")
    return targets_json.get("Results", [])


def _assign_labels(
    tokens: list[str], results: list[dict], positions: list[int], text: str
) -> np.ndarray:
    """Assign NER tags to tokens based on the results.

    Args:
        tokens: The tokens to assign labels to.
        results: The results to assign labels from.
        positions: The positions of the tokens in the text.
        text: The text.

    Returns:
        The assigned labels.
    """
    labels = np.array(["O"] * len(tokens), dtype=object)
    if not results:
        return labels

    for i, (token, token_start) in enumerate(zip(tokens, positions)):
        token_end = token_start + len(token)
        assert token == text[token_start:token_end], (
            f"Token {token} does not match text {text[token_start:token_end]}"
        )

        for entity in results:
            if token_start >= entity["Start"] and token_end <= entity["End"]:
                if token_start == entity["Start"]:
                    labels[i] = f"B-{entity['TypeName']}"
                else:
                    labels[i] = f"I-{entity['TypeName']}"

    return labels


def _get_positions(text: str) -> list[int]:
    """Get starting positions of tokens, including punctuation and words after spaces.

    Args:
        text: The text to get positions from.

    Returns:
        The starting positions of the tokens.
    """
    positions = []
    is_space_or_end = True

    for i, char in enumerate(text):
        if char.isspace():
            is_space_or_end = True
        elif re.match(r"[^\w\s]", char):
            # Punctuation: record the position
            positions.append(i)
            is_space_or_end = True
        elif is_space_or_end:
            # New word: record the position after a space
            positions.append(i)
            is_space_or_end = False

    return positions


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the KPWr NER dataset and upload it to the HF Hub."""

import pandas as pd
from datasets import Dataset, DatasetDict, Split, load_dataset
from huggingface_hub import HfApi


def main() -> None:
    """Create the KPWr NER dataset and uploads it to the HF Hub."""
    # Define dataset ID
    repo_id = "clarin-pl/kpwr-ner"

    # Download the dataset
    dataset = load_dataset(path=repo_id, token=True)
    assert isinstance(dataset, DatasetDict)

    # Convert the dataset to dataframes
    train_df = dataset["train"].to_pandas()
    test_df = dataset["test"].to_pandas()
    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(test_df, pd.DataFrame)

    # Drop unnecessary columns (keep only tokens and ner)
    columns_to_drop = [col for col in train_df.columns if col not in ["tokens", "ner"]]
    train_df.drop(columns=columns_to_drop, inplace=True)
    test_df.drop(columns=columns_to_drop, inplace=True)

    # Add a `text` column by joining tokens
    train_df["text"] = train_df["tokens"].map(lambda tokens: " ".join(tokens))
    test_df["text"] = test_df["tokens"].map(lambda tokens: " ".join(tokens))

    # Rename `ner` to `labels`
    train_df.rename(columns={"ner": "labels"}, inplace=True)
    test_df.rename(columns={"ner": "labels"}, inplace=True)

    # Get label names and create mapping
    label_names = dataset["train"].features["ner"].feature.names
    label_mapping = create_label_mapping(label_names)

    # Convert the NER tags from IDs to strings using our mapping
    train_df["labels"] = train_df["labels"].map(
        lambda ner_tags: [label_mapping.get(ner_tag, "O") for ner_tag in ner_tags]
    )
    test_df["labels"] = test_df["labels"].map(
        lambda ner_tags: [label_mapping.get(ner_tag, "O") for ner_tag in ner_tags]
    )

    # Create validation split from train set
    val_size = 256
    val_df = train_df.sample(n=val_size, random_state=4242)

    # Update train split (remove validation samples)
    train_size = 1024
    remaining_train = train_df.drop(val_df.index.tolist())
    train_df = remaining_train.sample(
        n=min(train_size, len(remaining_train)), random_state=4242
    )

    # Create test split
    test_size = 2048
    test_df = test_df.sample(n=min(test_size, len(test_df)), random_state=4242)

    # Reset the index
    train_df = train_df.reset_index(drop=True)
    val_df = val_df.reset_index(drop=True)
    test_df = test_df.reset_index(drop=True)

    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(val_df, pd.DataFrame)
    assert isinstance(test_df, pd.DataFrame)

    # Collect datasets in a dataset dictionary
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(test_df, split=Split.TEST),
        }
    )

    # Create dataset ID
    dataset_id = "EuroEval/kpwr-ner"

    # Remove the dataset from Hugging Face Hub if it already exists
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)

    # Push the dataset to the Hugging Face Hub
    dataset.push_to_hub(dataset_id, private=True)


def create_label_mapping(label_names: list[str]) -> dict[int, str]:
    """Create mapping from KPWr labels to standard BIO labels.

    Args:
        label_names: The list of label names.

    Returns:
        The mapping from KPWr labels to standard BIO labels.
    """
    mapping: dict[int, str] = {}

    for i, label_name in enumerate(label_names):
        if label_name == "O":
            mapping[i] = "O"
        elif "nam_liv" in label_name:
            # Living beings (persons, characters, etc.)
            if label_name.startswith("B-"):
                mapping[i] = "B-PER"
            elif label_name.startswith("I-"):
                mapping[i] = "I-PER"
        elif "nam_loc" in label_name:
            # Locations (cities, countries, etc.)
            if label_name.startswith("B-"):
                mapping[i] = "B-LOC"
            elif label_name.startswith("I-"):
                mapping[i] = "I-LOC"
        elif "nam_org" in label_name:
            # Organizations (companies, institutions, etc.)
            if label_name.startswith("B-"):
                mapping[i] = "B-ORG"
            elif label_name.startswith("I-"):
                mapping[i] = "I-ORG"
        else:
            # Everything else (events, products, etc.)
            if label_name.startswith("B-"):
                mapping[i] = "B-MISC"
            elif label_name.startswith("I-"):
                mapping[i] = "I-MISC"
            else:
                mapping[i] = "O"  # Fallback

    return mapping


if __name__ == "__main__":
    main()



"""Create the MIM-GOLD-NER-mini NER dataset and upload it to the HF Hub."""

url = "https://repository.clarin.is/repository/xmlui/bitstream/handle/20.500.12537/230/MIM-GOLD-2_0.zip?sequence=15&isAllowed=y"



"""Tests for the `finetuning` module."""

import pytest
import torch
from transformers.trainer_utils import IntervalStrategy
from transformers.training_args import TrainingArguments

from euroeval.data_models import BenchmarkConfig, ModelConfig
from euroeval.enums import DataType
from euroeval.finetuning import get_training_args


class TestGetTrainingArgs:
    """Test that the `get_training_args` function works as expected."""

    def test_return_type(
        self, benchmark_config: BenchmarkConfig, model_config: ModelConfig
    ) -> None:
        """Test that the return type is correct."""
        args = get_training_args(
            benchmark_config=benchmark_config,
            model_config=model_config,
            iteration_idx=0,
            dtype=DataType.FP32,
            batch_size=None,
        )
        assert isinstance(args, TrainingArguments)

    @pytest.mark.parametrize(argnames=["batch_size"], argvalues=[(8,), (16,), (32,)])
    def test_both_batch_sizes_are_set(
        self,
        benchmark_config: BenchmarkConfig,
        model_config: ModelConfig,
        batch_size: int,
    ) -> None:
        """Test that both training and eval batch sizes are set."""
        args = get_training_args(
            benchmark_config=benchmark_config,
            model_config=model_config,
            iteration_idx=0,
            dtype=DataType.FP32,
            batch_size=batch_size,
        )
        assert args.per_device_train_batch_size == batch_size
        assert args.per_device_eval_batch_size == batch_size

    @pytest.mark.parametrize(
        argnames=["batch_size", "expected_gradient_accumulation_steps"],
        argvalues=[(1, 32), (2, 16), (4, 8), (8, 4), (16, 2), (32, 1)],
    )
    def test_gradient_accumulation(
        self,
        benchmark_config: BenchmarkConfig,
        model_config: ModelConfig,
        batch_size: int,
        expected_gradient_accumulation_steps: int,
    ) -> None:
        """Test that the gradient accumulation is correct."""
        args = get_training_args(
            benchmark_config=benchmark_config,
            model_config=model_config,
            iteration_idx=0,
            dtype=DataType.FP32,
            batch_size=batch_size,
        )
        assert args.gradient_accumulation_steps == expected_gradient_accumulation_steps

    def test_batch_size_default_value(
        self, benchmark_config: BenchmarkConfig, model_config: ModelConfig
    ) -> None:
        """Test that the default value for the batch size is correct."""
        args = get_training_args(
            benchmark_config=benchmark_config,
            model_config=model_config,
            iteration_idx=0,
            dtype=DataType.FP32,
            batch_size=None,
        )
        assert (
            args.per_device_train_batch_size == benchmark_config.finetuning_batch_size
        )

    @pytest.mark.parametrize(
        argnames=["verbose", "expected_logging_strategy"],
        argvalues=[(True, IntervalStrategy.STEPS), (False, IntervalStrategy.NO)],
    )
    def test_logging_strategy(
        self,
        benchmark_config: BenchmarkConfig,
        model_config: ModelConfig,
        verbose: bool,
        expected_logging_strategy: IntervalStrategy,
    ) -> None:
        """Test that the logging strategy is correct."""
        old_verbose = benchmark_config.verbose
        benchmark_config.verbose = verbose
        args = get_training_args(
            benchmark_config=benchmark_config,
            model_config=model_config,
            iteration_idx=0,
            dtype=DataType.FP32,
            batch_size=None,
        )
        assert args.logging_strategy == expected_logging_strategy
        benchmark_config.verbose = old_verbose

    @pytest.mark.parametrize(
        argnames=["progress_bar", "expected_disable_tqdm"],
        argvalues=[(True, False), (False, True)],
    )
    def test_disable_tqdm(
        self,
        benchmark_config: BenchmarkConfig,
        model_config: ModelConfig,
        progress_bar: bool,
        expected_disable_tqdm: bool,
    ) -> None:
        """Test that the disable tqdm option is correct."""
        old_progress_bar = benchmark_config.progress_bar
        benchmark_config.progress_bar = progress_bar
        args = get_training_args(
            benchmark_config=benchmark_config,
            model_config=model_config,
            iteration_idx=0,
            dtype=DataType.FP32,
            batch_size=None,
        )
        assert args.disable_tqdm == expected_disable_tqdm
        benchmark_config.progress_bar = old_progress_bar

    @pytest.mark.parametrize(
        argnames=["datatype", "expected_fp16", "expected_bf16"],
        argvalues=[
            (DataType.FP32, False, False),
            (DataType.FP16, True, False),
            (DataType.BF16, False, True),
        ],
    )
    def test_dtype(
        self,
        benchmark_config: BenchmarkConfig,
        model_config: ModelConfig,
        datatype: DataType,
        expected_fp16: bool,
        expected_bf16: bool,
    ) -> None:
        """Test that the fp16 and bf16 arguments have been correctly set."""
        args = get_training_args(
            benchmark_config=benchmark_config,
            model_config=model_config,
            iteration_idx=0,
            dtype=datatype,
            batch_size=None,
        )
        assert args.fp16 == expected_fp16
        assert args.bf16 == expected_bf16

    @pytest.mark.parametrize(
        argnames=["device_name", "expected_use_cpu"],
        argvalues=[("cuda", False), ("mps", False), ("cpu", True)],
    )
    def test_use_cpu(
        self,
        benchmark_config: BenchmarkConfig,
        model_config: ModelConfig,
        device_name: str,
        expected_use_cpu: bool,
    ) -> None:
        """Test that the use_cpu argument is correct."""
        old_device = benchmark_config.device
        benchmark_config.device = torch.device(device_name)  # type: ignore[read-only]
        args = get_training_args(
            benchmark_config=benchmark_config,
            model_config=model_config,
            iteration_idx=0,
            dtype=DataType.FP32,
            batch_size=None,
        )
        assert args.use_cpu == expected_use_cpu
        benchmark_config.device = old_device  # type: ignore[read-only]



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create and upload the FullStack-LV-mini NER dataset from CoNLL-U format."""

import glob
import logging
import os
import shutil
import subprocess
from collections import defaultdict
from pathlib import Path

import pandas as pd
from datasets import Dataset, DatasetDict, Split
from huggingface_hub import HfApi

logger = logging.getLogger(__name__)


def main() -> None:
    """Create the FullStack-LV-mini NER dataset and upload it to the HF Hub."""
    # Clone the FullStack repository
    repo_path = clone_fullstack_repository()

    # Load and parse the data
    records = load_fullstack_data(repo_path=repo_path)

    # Convert to DataFrame
    df = pd.DataFrame.from_records(records)

    # Map the labels to the standard NER tags
    label_mapping = {
        "O": "O",
        "B-person": "B-PER",
        "I-person": "I-PER",
        "B-organization": "B-ORG",
        "I-organization": "I-ORG",
        "B-location": "B-LOC",
        "I-location": "I-LOC",
        "B-GPE": "B-LOC",  # Geopolitical entity -> Location
        "I-GPE": "I-LOC",
        "B-entity": "B-MISC",  # Generic entity -> Miscellaneous
        "I-entity": "I-MISC",
        "B-event": "B-MISC",  # Event -> Miscellaneous
        "I-event": "I-MISC",
        "B-product": "B-MISC",  # Product -> Miscellaneous
        "I-product": "I-MISC",
        "B-money": "B-MISC",  # Money -> Miscellaneous
        "I-money": "I-MISC",
        "B-time": "B-MISC",  # Time -> Miscellaneous
        "I-time": "I-MISC",
    }
    df["labels"] = df["labels"].map(
        lambda ner_tags: [label_mapping[ner_tag] for ner_tag in ner_tags]
    )

    # Create splits
    val_size = 256
    test_size = 2048
    train_size = 1024

    val_df = df.sample(n=val_size, random_state=4242)
    df_filtered = df[~df.index.isin(val_df.index)]

    test_df = df_filtered.sample(n=test_size, random_state=4242)
    df_filtered = df_filtered[~df_filtered.index.isin(test_df.index)]
    assert isinstance(df_filtered, pd.DataFrame)

    train_df = df_filtered.sample(n=train_size, random_state=4242)

    # Reset indices
    train_df = train_df.reset_index(drop=True)
    val_df = val_df.reset_index(drop=True)
    test_df = test_df.reset_index(drop=True)

    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(val_df, pd.DataFrame)
    assert isinstance(test_df, pd.DataFrame)

    # Collect datasets in a dataset dictionary
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(test_df, split=Split.TEST),
        }
    )

    # Create dataset ID
    dataset_id = "EuroEval/fullstack-ner-lv-mini"

    # Remove the dataset from Hugging Face Hub if it already exists
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)

    # Push the dataset to the Hugging Face Hub
    dataset.push_to_hub(dataset_id, private=True)

    # Delete the repository
    delete_fullstack_repository(repo_path=repo_path)


def clone_fullstack_repository(repo_name: str = "FullStack") -> Path:
    """Clone the FullStack repository if it doesn't already exist.

    Args:
        repo_name:
            Name of the directory to clone into.

    Returns:
        Path to the cloned repository.
    """
    if not Path(repo_name).exists():
        print("Cloning FullStack repository...")
        try:
            subprocess.run(
                [
                    "git",
                    "clone",
                    "https://github.com/LUMII-AILab/FullStack.git",
                    repo_name,
                ],
                check=True,
                capture_output=True,
                text=True,
            )
            logger.info("Successfully cloned repository")
        except subprocess.CalledProcessError as e:
            raise RuntimeError(f"Failed to clone repository: {e.stderr}") from e

    else:
        logger.info(f"Repository '{repo_name}' already exists, using existing copy")

    return Path(repo_name)


def load_fullstack_data(repo_path: Path) -> list[dict[str, list[str] | str]]:
    """Load and parse all FullStack NER data from the specified repository path.

    Args:
        repo_path:
            Path to the FullStack repository directory.

    Returns:
        A list of sentence records.
    """
    # Path to the data directory within the repository
    data_dir = repo_path / "NamedEntities" / "data"

    if not data_dir.exists():
        raise FileNotFoundError(
            f"Data directory '{data_dir}' not found in the repository."
        )

    # Find all .conll2003 files in the directory
    conll_files = glob.glob(os.path.join(data_dir, "*.conll2003"))

    all_records = []

    for file_path in sorted(conll_files):
        # Read the file
        with open(file_path, "r", encoding="utf-8") as f:
            content = f.read()

        # Parse the CoNLL-U data
        records = parse_conllu_data(content)
        all_records.extend(records)

    return all_records


def parse_conllu_data(raw_data: str) -> list[dict[str, list[str] | str]]:
    """Parse CoNLL-U format data and return a list of sentence records.

    Args:
        raw_data: The raw data in CoNLL-U format.

    Returns:
        A list of sentence records.
    """
    records = []
    lines = raw_data.strip().split("\n")

    # Initialize data dictionary for current sentence
    data_dict: dict[str, list[str]] = defaultdict(list)  # type: ignore[assignment]

    for line in lines:
        line = line.strip()

        # Skip comments (lines starting with #)
        if line.startswith("#"):
            continue

        # Empty line indicates end of sentence
        elif line == "":
            if len(data_dict["tokens"]) > 0:
                # Create record with tokens, labels, and text
                record = {
                    "tokens": data_dict["tokens"],
                    "labels": data_dict["labels"],
                    "text": " ".join(data_dict["tokens"]),
                }
                records.append(record)
            # Reset for next sentence
            data_dict = defaultdict(list)

        # Parse token line
        else:
            columns = line.split("\t")
            if len(columns) >= 7:  # Ensure we have enough columns
                token_id = columns[0]
                token = columns[1]
                # Use column 6 for NER tags (0-indexed, so columns[6] is column 7)
                ner_tag = columns[6] if len(columns) > 6 else "O"

                # Skip multi-word tokens (those with "-" in ID)
                if "-" not in token_id and "." not in token_id:
                    data_dict["tokens"].append(token)
                    data_dict["labels"].append(ner_tag)

    # Handle last sentence if data doesn't end with empty line
    if len(data_dict["tokens"]) > 0:
        record = {
            "tokens": data_dict["tokens"],
            "labels": data_dict["labels"],
            "text": " ".join(data_dict["tokens"]),
        }
        records.append(record)

    return records  # type: ignore[return-value]


def delete_fullstack_repository(repo_path: Path) -> None:
    """Delete the FullStack repository.

    Args:
        repo_path:
            Path to the FullStack repository.
    """
    if repo_path.exists():
        shutil.rmtree(repo_path)


if __name__ == "__main__":
    main()



"""All Finnish dataset configurations used in EuroEval."""

from ..data_models import DatasetConfig
from ..languages import FINNISH
from ..tasks import COMMON_SENSE, EUROPEAN_VALUES, LA, MCRC, NER, RC, SENT, SUMM

### Official datasets ###

SCANDISENT_FI_CONFIG = DatasetConfig(
    name="scandisent-fi",
    pretty_name="ScandiSent-fi",
    source="EuroEval/scandisent-fi-mini",
    task=SENT,
    languages=[FINNISH],
    _labels=["negative", "positive"],
)

TURKU_NER_FI_CONFIG = DatasetConfig(
    name="turku-ner-fi",
    pretty_name="Turku NER-fi",
    source="EuroEval/turku-ner-fi-mini",
    task=NER,
    languages=[FINNISH],
)

TYDIQA_FI_CONFIG = DatasetConfig(
    name="tydiqa-fi",
    pretty_name="TyDiQA-fi",
    source="EuroEval/tydiqa-fi-mini",
    task=RC,
    languages=[FINNISH],
)

XLSUM_FI_CONFIG = DatasetConfig(
    name="xlsum-fi",
    pretty_name="XLSum-fi",
    source="EuroEval/xlsum-fi-mini",
    task=SUMM,
    languages=[FINNISH],
)

HELLASWAG_FI_CONFIG = DatasetConfig(
    name="hellaswag-fi",
    pretty_name="HellaSwag-fi",
    source="EuroEval/hellaswag-fi-mini",
    task=COMMON_SENSE,
    languages=[FINNISH],
)

SCALA_FI_CONFIG = DatasetConfig(
    name="scala-fi",
    pretty_name="ScaLA-fi",
    source="EuroEval/scala-fi",
    task=LA,
    languages=[FINNISH],
)

VALEU_FI_CONFIG = DatasetConfig(
    name="valeu-fi",
    pretty_name="VaLEU-fi",
    source="EuroEval/european-values-fi",
    task=EUROPEAN_VALUES,
    languages=[FINNISH],
    splits=["test"],
    bootstrap_samples=False,
    _instruction_prompt="{text}",
)


### Unofficial datasets ###

BELEBELE_FI_CONFIG = DatasetConfig(
    name="belebele-fi",
    pretty_name="Belebele-fi",
    source="EuroEval/belebele-fi-mini",
    task=MCRC,
    languages=[FINNISH],
    unofficial=True,
)

MULTI_WIKI_QA_FI_CONFIG = DatasetConfig(
    name="multi-wiki-qa-fi",
    pretty_name="MultiWikiQA-fi",
    source="EuroEval/multi-wiki-qa-fi-mini",
    task=RC,
    languages=[FINNISH],
    unofficial=True,
)

GOLDENSWAG_FI_CONFIG = DatasetConfig(
    name="goldenswag-fi",
    pretty_name="GoldenSwag-fi",
    source="EuroEval/goldenswag-fi-mini",
    task=COMMON_SENSE,
    languages=[FINNISH],
    unofficial=True,
)

WINOGRANDE_FI_CONFIG = DatasetConfig(
    name="winogrande-fi",
    pretty_name="Winogrande-fi",
    source="EuroEval/winogrande-fi",
    task=COMMON_SENSE,
    languages=[FINNISH],
    _labels=["a", "b"],
    unofficial=True,
)



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
#     "scikit-learn<1.6.0",
# ]
# ///

"""Create the CS GEC linguistic acceptability dataset."""

import logging

import pandas as pd
from datasets import Dataset, DatasetDict, Split, load_dataset
from huggingface_hub import HfApi
from sklearn.model_selection import train_test_split

from .constants import MAX_NUM_CHARS_IN_DOCUMENT, MIN_NUM_CHARS_IN_DOCUMENT  # noqa

logging.basicConfig(format="%(asctime)s ⋅ %(message)s", level=logging.INFO)
logger = logging.getLogger("create_cs_gec")


def main() -> None:
    """Create the CS GEC linguistic acceptability dataset and upload to HF Hub."""
    # Load the dataset using Hugging Face datasets library
    dataset = load_dataset("CZLC/cs_gec")
    assert isinstance(dataset, DatasetDict)

    # Convert the train and test datasets to pandas DataFrames
    train_df = dataset["train"].to_pandas()
    test_df = dataset["test"].to_pandas()
    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(test_df, pd.DataFrame)

    # Rename columns to match the expected format
    train_df = train_df.rename(columns={"query": "text", "gold": "label"})
    test_df = test_df.rename(columns={"query": "text", "gold": "label"})
    train_df["label"] = train_df["label"].apply(
        lambda x: "correct" if x == 1 else "incorrect"
    )
    test_df["label"] = test_df["label"].apply(
        lambda x: "correct" if x == 1 else "incorrect"
    )

    # Filter data based on text length
    train_df = filter_by_text_length(train_df)
    test_df = filter_by_text_length(test_df)

    # Sample a larger number of rows to create train and val splits
    sampled_train_df = train_df.sample(n=1280, random_state=42)  # 1024 + 256

    # Create a validation split from the sampled train data
    train_df, val_df = train_test_split(
        sampled_train_df, train_size=1024, random_state=42
    )

    # Sample the desired number of rows for the test split
    test_df = test_df.sample(n=2048, random_state=42)

    # Create a DatasetDict
    dataset_dict = DatasetDict(
        {
            "train": Dataset.from_pandas(
                train_df, split=Split.TRAIN, preserve_index=False
            ),
            "val": Dataset.from_pandas(
                val_df, split=Split.VALIDATION, preserve_index=False
            ),
            "test": Dataset.from_pandas(
                test_df, split=Split.TEST, preserve_index=False
            ),
        }
    )

    # Upload to Hugging Face Hub
    dataset_id = "EuroEval/cs-gec-mini"
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)
    dataset_dict.push_to_hub(dataset_id, private=True)


def filter_by_text_length(df: pd.DataFrame) -> pd.DataFrame:
    """Filter the dataframe by text length."""
    df["text_len"] = df.text.str.len()
    return df.query(
        "text_len >= @MIN_NUM_CHARS_IN_DOCUMENT and "
        "text_len <= @MAX_NUM_CHARS_IN_DOCUMENT"
    )


if __name__ == "__main__":
    main()
